# RCAIDE/Library/Methods/Geometry/Two_Dimensional/Airfoil/convert_airfoil_to_meshgrid.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# convert_airfoil_to_meshgrid
# ----------------------------------------------------------------------------------------------------------------------   
def convert_airfoil_to_meshgrid(airfoil_geometry, *args, **kwargs):
    """Converts a RCAIDE airfoil geometry representation to a Numpy meshgrid
    array mask of boolean values.

    Assumptions:
    None

    Source:
    None

    Inputs:
    airfoil_geometry                [RCAIDE Data Structure]
        .x_lower_surface            [Numpy Array, float32]
        .y_lower_surface            [Numpy Array, float32]
        .y_upper_surface            [Numpy Array, float32]

    Outputs: 

    airfoil_meshgrid                [Numpy Array, bool]

    Properties Used:
    N/A	
    """

    # Unpack Values

    x_lower_surface = airfoil_geometry.x_lower_surface
    y_lower_surface = airfoil_geometry.y_lower_surface
    y_upper_surface = airfoil_geometry.y_upper_surface

    # Determine necessary resolution of the meshgrid. We do this by dividing the
    # x-length of the  lower surface by the minimum separation between
    # any two x-coordinates of the geometry (ceil-rounded to an int). Later
    # we'll instantiate the meshgrid with this number of x-indices, so that the
    # separation between any two points in the meshgrid is equal to the minimum
    # separation between any two x-coordinates.

    x_length = (
        np.max(x_lower_surface)
    )

    Nx = np.ceil(
        x_length / np.abs(np.min(np.diff(x_lower_surface)))
    ).astype(int)

    # We determine the necessary number of y-coordinate points by taking the
    # maximum separation between the highest point of the upper surface and the
    # lowest point of the lower surface and multiplying that by the number of
    # x-points in order to re-normalize to our future meshgrid coordinates,
    # then ciel-rounding to an int.

    Ny = np.ceil(
        Nx * ( np.max(y_upper_surface) - np.min(y_lower_surface) )
    ).astype(int)

    # Instantiate the meshgrid, using ij-indexing so that X[i,j] returns i
    # for all points, and Y[i,j] returns j for all coordinates.

    X, Y = np.meshgrid(np.arange(Nx), np.arange(Ny), indexing="ij")

    # Create the indexing arrays for the meshgrid. These convert the airfoil
    # geometry coordinates into meshgrid array indices. The X_INDICES are found
    # just by multplying/stretching the x_lower_surface coordinates across the
    # number of x-coodinates in the meshgrid.

    X_INDICES = np.ceil(
        Nx / x_length * x_lower_surface
    ).astype(int)

    # The Y_INDICES are similarly stretched, but first are offset by the
    # minimum of the lower surface to bring them to a relative zero

    Y_LOWER_INDICES = np.floor(
        Nx / x_length * (
            y_lower_surface - np.min(y_lower_surface)
        )
    ).astype(int)

    Y_UPPER_INDICES = np.ceil(
        Nx /x_length * (
            y_upper_surface - np.min(y_lower_surface)
        )
    ).astype(int)

    # We then repeat the elements of the Y_INDICES by the number of gridpoints
    # between each x-coordinate, essentially treating the y-surface as flat
    # between those points. We trim the final point by telling it to repeat 0
    # times

    REPEATS = np.append(
            np.diff(X_INDICES),
            0
    )

    # Need to hand the case where the X_INDICES aren't sorted, and swap
    # some elements around to allow the masks to be created

    if np.any(REPEATS<0):

        REPEAT_FLAG = True

        NEG_REPEATS = np.where(REPEATS<0)[0]

        if np.any(np.diff(NEG_REPEATS) == 1):
            print("Airfoil geometry contains sequential negative x-steps. Meshing Failed.")
            return None

        (X_INDICES[NEG_REPEATS],
         X_INDICES[NEG_REPEATS + 1]) = (X_INDICES[NEG_REPEATS + 1],
                                        X_INDICES[NEG_REPEATS])

        (Y_LOWER_INDICES[NEG_REPEATS],
         Y_LOWER_INDICES[NEG_REPEATS + 1]) = (Y_LOWER_INDICES[NEG_REPEATS + 1],
                                              Y_LOWER_INDICES[NEG_REPEATS])

        (Y_UPPER_INDICES[NEG_REPEATS],
         Y_UPPER_INDICES[NEG_REPEATS + 1]) = (Y_UPPER_INDICES[NEG_REPEATS + 1],
                                              Y_UPPER_INDICES[NEG_REPEATS])

        REPEATS = np.append(
            np.diff(X_INDICES),
            0
        )

        Nx = np.sum(REPEATS)

        Ny = np.ceil(
            Nx * (np.max(y_upper_surface) - np.min(y_lower_surface))
        ).astype(int)

        X, Y = np.meshgrid(np.arange(Nx), np.arange(Ny), indexing="ij")

    Y_LOWER_INDICES = np.repeat(Y_LOWER_INDICES, REPEATS)
    Y_UPPER_INDICES = np.repeat(Y_UPPER_INDICES, REPEATS)

    # We then create masks for the upper and lower surfaces by tiling the
    # indices over the meshgrid (taking a transpose to comport with our earlier
    # indexing style).

    Y_LOWER_GRID = np.tile(Y_LOWER_INDICES, (Ny,1)).T
    Y_UPPER_GRID = np.tile(Y_UPPER_INDICES, (Ny,1)).T

    # We then create our airfoil meshgrid mask by comparing our Y coordinates
    # from the meshgrid to our upper and lower grids, intermediately treating
    # them as ints to simplify the multi-condition comparison

    Y_LOWER = (Y > Y_LOWER_GRID).astype(int)
    Y_UPPER = (Y < Y_UPPER_GRID).astype(int)

    AIRFOIL_MASK = (Y_LOWER + Y_UPPER) > 1

    return AIRFOIL_MASK



# RCAIDE/Library/Attributes/Solids/Polyetherimide.py
# 
# 
#
# Created: Sep 2024 S. Shekar

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Polyetherimide for Reservoir Casing
#-------------------------------------------------------------------------------
class Polyetherimide(Solid):

    """ Physical Constants Specific to Polyetherimide   
    """
    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        https://www.matweb.com/search/DataSheet.aspx?MatGUID=65baf7a4f90c4a54a6ace03e16b1125b&amp%3bckck=1&ckck=1

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """ 
        self.conductivity              = 0.2    # [W/m-K]
        self.emissivity                = 0.96   # [uniteless]
        self.specific_heat             = 1100  



# RCAIDE/Library/Missions/Segments/Cruise/Constant_Acceleration_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant acceleration and constant altitude

    Source:
    N/A

    Inputs:
    segment.altitude                [meters]
    segment.air_speed_start         [meters/second]
    segment.air_speed_end           [meters/second]
    segment.acceleration            [meters/second^2]
    conditions.frames.inertial.time [seconds]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    alt        = segment.altitude 
    v0         = segment.air_speed_start
    vf         = segment.air_speed_end
    ax         = segment.acceleration    
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions 
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for initial velocity
    if v0 is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        v0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = (vf-v0)/ax + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial 
    v_mag = v0+(time - t_initial)*ax
    v_x   = np.cos(beta)*v_mag
    v_y   = np.sin(beta)*v_mag
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0] 



# RCAIDE/Framework/Analyses/Mission/Segments/Conditions/State.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import DataOrdered
from .Conditions           import Conditions
from .Unknowns             import Unknowns
from .Residuals            import Residuals
from .Numerics             import Numerics   

# python imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  State
# ----------------------------------------------------------------------------------------------------------------------
class State(Conditions):
    """ Creates the State data structure for storing daata that solved in a mission
    
        Assumptions:
        None
        
        Source:
        None
    """    
    
    
    def __defaults__(self):
        """ This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        self.tag        = 'state'
        self.initials   = Conditions()
        self.numerics   = Numerics()
        self.unknowns   = Unknowns()
        self.conditions = Conditions()
        self.residuals  = Residuals()
        
    def expand_rows(self,rows,override=False):
        """ Makes a 1-D array the right size. Often used after a mission is initialized to size out the vectors to the
            right size. Will not overwrite an array if it already exists, unless override is True.
        
            Assumptions:
            Doesn't expand initials or numerics
    
            Source:
            N/A
    
            Inputs:
            rows   [int]
    
            Outputs:
            None
    
            Properties Used:
            None
        """         
        
        # store
        self._size = rows
        
        for k,v in self.items(): 
            try:
                rank = v.ndim
            except:
                rank = 0            
            # don't expand initials or numerics
            if k in ('initials','numerics'):
                continue
            
            # recursion
            elif isinstance(v,Conditions):
                v.expand_rows(rows,override=override)
            # need arrays here
            elif rank == 2:
                self[k] = np.resize(v,[rows,v.shape[1]])
            #: if type
        #: for each key,value        
        
# ----------------------------------------------------------------------------------------------------------------------
# Container
# ----------------------------------------------------------------------------------------------------------------------        
                
class Container(State):
    def __defaults__(self):
        """ This sets the default values.
    
            Assumptions:
            Puts the segments in the right order
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """         
        self.segments = DataOrdered()
        
    def merged(self):
        """ Combines the states of multiple segments
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            state_out [State()]
    
            Properties Used:
            None
        """              
        
        state_out = State()
        
        for i,(tag,sub_state) in enumerate(self.segments.items()):
            for key in ['unknowns','conditions','residuals']:
                if i == 0:
                    state_out[key].update(sub_state[key])
                else:
                    state_out[key] = state_out[key].do_recursive(append_array,sub_state[key])
            
        return state_out
        
State.Container = Container 
        
# ----------------------------------------------------------------------------------------------------------------------
# append_array
# ---------------------------------------------------------------------------------------------------------------------- 

def append_array(A,B=None):
    """ A stacking operation used by merged to put together data structures

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        A [array]
        B [array]

        Outputs:
        array

        Properties Used:
        None
    """       
    if isinstance(A,np.ndarray) and isinstance(B,np.ndarray):
        return np.vstack([A,B])
    else:
        return None


# RCAIDE/Library/Missions/Common/Initialize/differentials_dimensionless.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports 
from RCAIDE.Framework.Core.Arrays  import atleast_2d_col 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Differentials
# ----------------------------------------------------------------------------------------------------------------------
def differentials_dimensionless(segment):
    """ Discretizes the differential operators
    
        Assumptions:
        N/A
        
        Inputs:
            state.numerics:
                number_of_control_points [int]
                discretization_method    [function]
            
        Outputs:
            numerics.dimensionless:            
                control_points           [array]
                differentiate            [array]
                integrate                [array]

        Properties Used:
        N/A
                                
    """     
    
    # unpack
    numerics              = segment.state.numerics
    N                     = numerics.number_of_control_points
    discretization_method = numerics.discretization_method
    
    # get operators
    x,D,I = discretization_method(N,**numerics)
    x = atleast_2d_col(x)
    
    # pack
    numerics.dimensionless.control_points = x
    numerics.dimensionless.differentiate  = D
    numerics.dimensionless.integrate      = I    
    
    return
 


# RCAIDE/Framework/External_Interfaces/OpenVSP/export_vsp_vehicle.py
# 
# Created:  Jul 2016, T. MacDonald
# Modified: Jun 2017, T. MacDonald
#           Jul 2017, T. MacDonald
#           Oct 2018, T. MacDonald
#           Nov 2018, T. MacDonald
#           Jan 2019, T. MacDonald
#           Jan 2020, T. MacDonald 
#           Mar 2020, M. Clarke
#           May 2020, E. Botero
#           Jul 2020, E. Botero 
#           Feb 2021, T. MacDonald
#           May 2021, E. Botero 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units, Data 
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_rotor     import write_vsp_rotor_bem
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_fuselage  import write_vsp_fuselage
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_wing      import write_vsp_wing
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_nacelle   import write_vsp_nacelle 
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
import numpy as np
import os

# ---------------------------------------------------------------------------------------------------------------------- 
# export_vsp_vehicle
# ---------------------------------------------------------------------------------------------------------------------- 
def export_vsp_vehicle(vehicle, vehicle_tag, fuel_tank_set_ind=3, verbose=True, write_file=True, OML_set_ind = 4, write_igs = False):
    """This writes a RCAIDE vehicle to OpenVSP format. It will take wing segments into account
    if they are specified in the vehicle setup file.
    
    Assumptions:
    Vehicle is composed of conventional shape fuselages, wings, and networks. Any network
    that should be created is tagged as 'turbofan'.

    Source:
    N/A

    Inputs:
    vehicle.
      tag                                       [-]
      wings.*.    (* is all keys)
        origin                                  [m] in all three dimensions
        spans.projected                         [m]
        chords.root                             [m]
        chords.tip                              [m]
        sweeps.quarter_chord                    [radians]
        twists.root                             [radians]
        twists.tip                              [radians]
        thickness_to_chord                      [-]
        dihedral                                [radians]
        tag                                     <string>
        Segments.*. (optional)
          twist                                 [radians]
          percent_span_location                 [-]  .1 is 10%
          root_chord_percent                    [-]  .1 is 10%
          dihedral_outboard                     [radians]
          sweeps.quarter_chord                  [radians]
          thickness_to_chord                    [-]
      networks.turbofan. (optional)
        number_of_engines                       [-]
        nacelle_diameter                        [m]
        origin                                  [m] in all three dimension, should have as many origins as engines
        OpenVSP_simple (optional)               <boolean> if False (default) create a flow through nacelle, if True creates a roughly biparabolic shape
      fuselages.fuselage (optional)
        width                                   [m]
        lengths.total                           [m]
        heights.
          maximum                               [m]
          at_quarter_length                     [m]
          at_wing_root_quarter_chord            [m]
          at_three_quarters_length              [m]
        effective_diameter                      [m]
        fineness.nose                           [-] ratio of nose section length to fuselage width
        fineness.tail                           [-] ratio of tail section length to fuselage width
        tag                                     <string>
        OpenVSP_values.  (optional)
          nose.top.angle                        [degrees]
          nose.top.strength                     [-] this determines how much the specified angle influences that shape
          nose.side.angle                       [degrees]
          nose.side.strength                    [-]
          nose.TB_Sym                           <boolean> determines if top angle is mirrored on bottom
          nose.z_pos                            [-] z position of the nose as a percentage of fuselage length (.1 is 10%)
          tail.top.angle                        [degrees]
          tail.top.strength                     [-]
          tail.z_pos (optional, 0.02 default)   [-] z position of the tail as a percentage of fuselage length (.1 is 10%)
    fuel_tank_set_index                         <int> OpenVSP object set containing the fuel tanks    

    Outputs:
    <tag>.vsp3           This is the OpenVSP representation of the aircraft

    Properties Used:
    N/A
    """    

    # -------------------------------------------------------------------------     
    # Reset OpenVSP to avoid including a previous vehicle
    # ------------------------------------------------------------------------- 
    if verbose:
        print('Reseting OpenVSP Model in Memory')
    try:
        vsp.ClearVSPModel()
    except NameError:
        print('VSP import failed')
        return -1
    
    area_tags = dict() # for wetted area assignment
    
    # ------------------------------------------------------------------------- 
    # Wings
    # -------------------------------------------------------------------------  
    # Default Set_0 in OpenVSP is index 3
    vsp.SetSetName(fuel_tank_set_ind, 'fuel_tanks')
    vsp.SetSetName(OML_set_ind, 'OML')
    
    for wing in vehicle.wings:       
        if verbose:
            print('Writing '+wing.tag+' to OpenVSP Model')
            area_tags, wing_id = write_vsp_wing(vehicle,wing,area_tags, fuel_tank_set_ind, OML_set_ind) 
    
    # ------------------------------------------------------------------------- 
    # Engines
    # -------------------------------------------------------------------------  
    for network in vehicle.networks: 
        for propulsor in network.propulsors: 
            for  tag ,  item in  propulsor.items():
                if isinstance(item, RCAIDE.Library.Components.Propulsors.Converters.Rotor):
                    vsp_bem_filename = item.tag + '.bem' 
                    write_vsp_rotor_bem(vsp_bem_filename,item) 
   
            if 'nacelle' in propulsor:
                nacelle =  propulsor.nacelle
                if verbose:
                    print('Writing '+ nacelle.tag +' to OpenVSP Model')
                write_vsp_nacelle(nacelle, OML_set_ind)
                     
    # ------------------------------------------------------------------------- 
    # Fuselage
    # ------------------------------------------------------------------------- 
    for fuselage in vehicle.fuselages:
        if type(fuselage) !=  RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage: 
            if verbose:
                print('Writing '+fuselage.tag+' to OpenVSP Model')
            try:
                area_tags = write_vsp_fuselage(fuselage, area_tags, vehicle.wings.main_wing, 
                                               fuel_tank_set_ind, OML_set_ind)
            except AttributeError:
                area_tags = write_vsp_fuselage(fuselage, area_tags, None, fuel_tank_set_ind,
                                               OML_set_ind)
        else:
            print('Blended Wing Body Fuselage defined')
    
    vsp.Update()
    

    # ------------------------------------------------------------------------- 
    # Fuselage
    # ------------------------------------------------------------------------- 
    for boom in vehicle.booms: 
        if verbose:
            print('Writing '+boom.tag+' to OpenVSP Model')
        try:
            area_tags = write_vsp_fuselage(boom, area_tags, vehicle.wings.main_wing, 
                                           fuel_tank_set_ind, OML_set_ind)
        except AttributeError:
            area_tags = write_vsp_fuselage(boom, area_tags, None, fuel_tank_set_ind,
                                               OML_set_ind) 
    # -------------------------------------------------------------------------     
    # Write the vehicle to the file    
    # ------------------------------------------------------------------------- 
    if write_file ==True:
        cwd = os.getcwd()
        filename = vehicle_tag + ".vsp3"
        if verbose:
            print('Saving OpenVSP File at '+ cwd + '/' + filename)
        vsp.WriteVSPFile(filename)
    elif verbose:
        print('Not Saving OpenVSP File')
        
    if write_igs:
        if verbose:
            print('Exporting IGS File')        
        vehicle_id = vsp.FindContainersWithName('Vehicle')[0]
        parm_id = vsp.FindParm(vehicle_id,'LabelID','IGESSettings')
        vsp.SetParmVal(parm_id, 0.)
        vsp.ExportFile(vehicle_tag + ".igs", OML_set_ind, vsp.EXPORT_IGES)
    
    return area_tags


## @ingroup Methods-Performance
# find_take_off_weight_given_tofl.py
#
# Created:  Sep 2014, C. Ilario, T. Orra 
# Modified: Jan 2016, E. Botero


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from RCAIDE.Library.Methods.Performance.estimate_take_off_field_length import estimate_take_off_field_length

import numpy as np

# ----------------------------------------------------------------------
#  Find Takeoff Weight Given TOFL
# ----------------------------------------------------------------------

## @ingroup Methods-Performance
def find_take_off_weight_given_tofl(vehicle,analyses,target_tofl,altitude = 0, delta_isa = 0,):
    """Estimates the takeoff weight given a certain takeoff field length.

    Assumptions:
    assumptions per estimate_take_off_field_length()

    Source:
    N/A

    Inputs:
    vehicle.mass_properties.
      operating_empty         [kg]
      max_takeoff             [kg]
      analyses                [RCAIDE data structure]
      airport                 [RCAIDE data structure]
      target_tofl             [m]
      
    Outputs:
    max_tow                   [kg]

    Properties Used:
    N/A
    """       

    #unpack
    tow_lower = vehicle.mass_properties.operating_empty
    tow_upper = 1.10 * vehicle.mass_properties.max_takeoff

    #saving initial reference takeoff weight
    tow_ref = vehicle.mass_properties.max_takeoff

    tow_vec = np.linspace(tow_lower,tow_upper,50)
    tofl    = np.zeros_like(tow_vec)

    for id,tow in enumerate(tow_vec):
        vehicle.mass_properties.takeoff = tow
        tofl[id], _ = estimate_take_off_field_length(vehicle,analyses,altitude = 0, delta_isa = 0)

    target_tofl = np.atleast_1d(target_tofl)
    max_tow     = np.zeros_like(target_tofl)

    for id,toflid in enumerate(target_tofl):
        max_tow[id] = np.interp(toflid,tofl,tow_vec)

    #reset the initial takeoff weight
    vehicle.mass_properties.max_takeoff = tow_ref

    return max_tow


# RCAIDE/Framework/Analyses/Propulsion/Ducted_Fan_Design_Code.py
#  
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
from RCAIDE.Framework.Core import  Data , Units, Container
from RCAIDE.Framework.Analyses import Analysis   

# package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
# Ducted_Fan_Design_Code
# ---------------------------------------------------------------------------------------------------------------------- 
class Ducted_Fan_Design_Code(Analysis): 
    def __defaults__(self):
        """This sets the default values and methods for the analysis.

        Assumptions:
            None

        Source:
            https://web.mit.edu/drela/Public/web/dfdc/
            http://www.esotec.org/sw/DFDC.html 
        """          
        self.tag                                          = 'dfdc'    
                                                          
        self.current_status                               = Data()        
        self.current_status.batch_index                   = 0
        self.current_status.batch_file                    = None
        self.current_status.deck_file                     = None
        self.current_status.cases                         = None      
        self.geometry                                     = None   
                                                          
        self.settings                                     = Data()
        self.settings.filenames                           = Data()
        self.settings.filenames.dfdc_bin_name             = 'dfdc'   # to call dfdc from command line. If avl is not on the system path, include absolute path to the avl binary i.e. '/your/path/to/avl'
        self.settings.filenames.run_folder                = 'dfdc_files'   
        self.settings.filenames.deck_template             = 'commands_{0:02d}.deck'  
        self.settings.filenames.results_template          = 'results_Vinf_{:.2f}_RPM_{:.2f}_Alt_{:.2f}' 
        self.settings.filenames.case                      = None
        self.settings.filenames.log_filename              = 'dfdc_log.txt'
        self.settings.filenames.err_filename              = 'dfdc_err.txt'
 
        self.settings.print_output                        = False 
                   
        # Regression Status           
        self.settings.keep_files                          = False           
        self.settings.new_regression_results              = False   
    
        self.run_cases                                    = Container() 
        
        # Conditions table, used for surrogate model training
        self.training                                     = Data()   
                  
        # Standard subsonic/transolic aircarft          
        self.training.tip_mach                            = np.array([0.2,0.5,0.8])     
        self.training.mach                                = np.array([0.1,0.3,0.6]) 
        self.training.altitude                            = np.array([0,1000,2500,5000])*Units.feet    
                                                        
        self.training_file                                = None
                  
        # Surrogate model          
        self.surrogates                                   = Data()  
 
     
    def append_case(self,case):
        """ Adds a case to the set of run cases "
        
	Assumptions:
	    None
    
	Source:
	    None
    
	Inputs:
	    None
    
	Outputs:
	    None
    
	Properties Used:
	    N/A
	"""
        run_cases   =  self.run_cases  
        run_cases.append(case) 
        return



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ---------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(vehicle, wing, rudder_fraction=0.25):
    """ Calculate the weight of the vertical fin of an aircraft without the weight of 
    the rudder and then calculate the weight of the rudder 
    
    Assumptions:
        Vertical tail weight is the weight of the vertical fin without the rudder weight.
        Rudder occupies 25% of the S_v and weighs 60% more per unit area.     
        
    Source: 
        N/A 
        
    Inputs:
        S_v - area of the vertical tail (combined fin and rudder)                      [meters**2]
        vehicle.flight_envelope.ultimate_load - ultimate load of the aircraft                 [dimensionless]
        wing.spans.projected - span of the vertical                                    [meters]
        vehicle.mass_properties.max_takeoff - maximum takeoff weight of the aircraft   [kilograms]
        wing.thickness_to_chord- thickness-to-chord ratio of the vertical tail         [dimensionless]
        wing.sweeps.quarter_chord - sweep angle of the vertical tail                   [radians]
        vehicle.reference_area - wing gross area                                       [meters**2]
        wing.t_tail - factor to determine if aircraft has a t-tail                     [dimensionless]
        rudder_fraction - fraction of the vertical tail that is the rudder             [dimensionless]
    
    Outputs:
        output - a dictionary with outputs:
            W_tail_vertical - weight of the vertical fin portion of the vertical tail [kilograms]
            W_rudder - weight of the rudder on the aircraft                           [kilograms]
  
    Properties Used:
        N/A
    """
    # unpack inputs
    span                   = wing.spans.projected / Units.ft  # Convert meters to ft
    sweep                  = wing.sweeps.quarter_chord  # Convert deg to radians
    area                   = wing.areas.reference / Units.ft ** 2  # Convert meters squared to ft squared
    mtow                   = vehicle.mass_properties.max_takeoff / Units.lb  # Convert kg to lbs
    Sref                   = vehicle.reference_area / Units.ft ** 2  # Convert from meters squared to ft squared
    thickness_to_chord_v   = wing.thickness_to_chord
    # Determine weight of the vertical portion of the tail
    if wing.t_tail == "yes":
        T_tail_factor = 1.25  # Weight of vertical portion of the T-tail is 25% more than a conventional tail
    else:
        T_tail_factor = 1.0

        # Calculate weight of wing for traditional aircraft vertical tail without rudder
    tail_vert_English = T_tail_factor * (
                2.62 * area + 1.5 * 10. ** (-5.) * vehicle.flight_envelope.ultimate_load * span ** 3. * (8. + 0.44 * mtow / Sref) / (
                    thickness_to_chord_v * (np.cos(sweep) ** 2.)))

    tail_weight  = tail_vert_English * Units.lbs
    tail_weight += tail_weight * rudder_fraction * 1.6

    return tail_weight



# RCAIDE/Methods/Aerodynamics/Common/Lift/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 


from .compute_max_lift_coeff                  import compute_max_lift_coeff
from .compute_flap_lift                       import compute_flap_lift
from .compute_slat_lift                       import compute_slat_lift 
from .BET_calculations                        import compute_airfoil_aerodynamics 
from .BET_calculations                        import compute_inflow_and_tip_loss
from .fuselage_correction                     import fuselage_correction
from .generate_wing_wake_grid                 import generate_wing_wake_grid


## @ingroup Methods-Utilities
# Cubic_Spline_Blender.py
# 
# Created:  Feb 2019, T. MacDonald
# Modified: Jan 2020, T. MacDonald (moved from Method/Aerodynamics/Supersonic_Zero/Drag)

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np

# ----------------------------------------------------------------------
#  Blender Class
# ----------------------------------------------------------------------

## @ingroup Methods-Utilities
class Cubic_Spline_Blender():
    """This is a cubic spline function that can be used to blend two type of calculations
    without knowing the end points. It preserves continuous first derivatives.

    Assumptions:
    None

    Source:
    Information at:
    https://en.wikipedia.org/wiki/Cubic_Hermite_spline
    """ 
    
    def __init__(self, x_start, x_end):
        """This sets the default start and end position.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """           
        self.x_start = x_start
        self.x_end   = x_end
    
    
    def compute(self,x):
        """This computes the y value along a normalized spline

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """          
        eta = self.eta_transform(x)
    
        y = 2*eta*eta*eta-3*eta*eta+1
        y[eta<0] = 1
        y[eta>1] = 0
        return y

    def eta_transform(self,x):
        """Normalizes the transformation

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """          
        x_start = self.x_start
        x_end   = self.x_end
        
        eta     = (x-x_start)/(x_end-x_start)
        
        return eta


# RCAIDE/Library/Methods/Propulsors/Converters/DC_Motor/compute_motor_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# compute_Q_from_omega_and_V
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_torque_from_RPM_and_voltage(motor,motor_conditions,conditions):
    """Calculates the motor's torque based on RPM (angular velocity) and voltage.  
    The following perperties of the motor are computed
    
    motor_conditions.torque  (numpy.ndarray): torque  [Nm]
    motor_conditions.omega   (numpy.ndarray): omega   [radian/s]
      
    Assumptions:
       None 

    Source:
       None

    Args:
        motor.
          gear_ratio                  (float): gear ratio                [unitless]
          speed_constant              (float): motor speed constant      [radian/s/V]
          resistance                  (float): motor internal resistnace [ohm]
          outputs.omega       (numpy.ndarray): angular velocity          [radian/s]
          gearbox_efficiency          (float): gearbox efficiency        [unitless]
          expected_current            (float): current                   [A]
          no_load_current             (float): no-load current           [A]
          inputs.volage       (numpy.ndarray): operating voltage         [V]

    Returns:
        None 
 
    """
    
    Res   = motor.resistance
    eta_G = motor.gearbox_efficiency
    exp_I = motor.expected_current
    I0    = motor.no_load_current + exp_I*(1-eta_G)
    G     = motor.gear_ratio
    KV    = motor.speed_constant/G
    v     = motor_conditions.voltage
    omega = motor_conditions.omega
    
    # Torque
    Q = ((v-omega/KV)/Res -I0)/KV
    
    motor_conditions.torque = Q
    motor_conditions.omega  = omega

    return


# ----------------------------------------------------------------------------------------------------------------------
#  compute_omega_and_Q_from_Cp_and_V
# ----------------------------------------------------------------------------------------------------------------------    
def compute_RPM_and_torque_from_power_coefficent_and_voltage(motor,motor_conditions,conditions):
    """Calculates the motors RPM and torque using power coefficient and operating voltage.
    The following perperties of the motor are computed  
    motor_conditions.torque                    (numpy.ndarray):  torque [Nm]
    motor_conditions.omega                     (numpy.ndarray):  omega  [radian/s] 

    Assumptions: 
      Omega is solved by setting the torque of the motor equal to the torque of the prop
      It assumes that the Cp is constant 

    Source:
        None

    Args:
    conditions.freestream.density  (numpy.ndarray): density [kg/m^3] 
    motor.
        gear_ratio                                 (float): gear ratio                [unitless]
        speed_constant                             (float): motor speed constant      [radian/s/V]
        resistance                                 (float): motor internal resistnace [ohm]
        outputs.omega                      (numpy.ndarray): angular velocity          [radian/s]
        gearbox_efficiency                         (float): gearbox efficiency        [unitless]
        expected_current                           (float): current                   [A]
        no_load_current                            (float): no-load current           [A]
        inputs.volage                      (numpy.ndarray): operating voltage         [V]
        inputs.rotor_power_coefficient     (numpy.ndarray): power coefficient         [unitless]
        rotor_radius                               (float): rotor radius              [m]
           
    Returns:
        None
    """           
    # Unpack 
    rho   = conditions.freestream.density[:,0,None]
    Res   = motor.resistance
    eta_G = motor.gearbox_efficiency
    exp_I = motor.expected_current
    I0    = motor.no_load_current + exp_I*(1-eta_G)
    G     = motor.gear_ratio
    KV    = motor.speed_constant/G
    R     = motor.rotor_radius
    v     = motor_conditions.voltage
    Cp    = motor_conditions.rotor_power_coefficient 

    # compute angular velocity, omega 
    omega   =   ((np.pi**(3./2.))*((- 16.*Cp*I0*rho*(KV*KV*KV)*(R*R*R*R*R)*(Res*Res) +
                16.*Cp*rho*v*(KV*KV*KV)*(R*R*R*R*R)*Res + (np.pi*np.pi*np.pi))**(0.5) - 
                np.pi**(3./2.)))/(8.*Cp*(KV*KV)*(R*R*R*R*R)*Res*rho)
    omega [np.isnan(omega )] = 0.0
    
    # compute torque 
    Q = ((v-omega /KV)/Res -I0)/KV 
    
    # store values 
    motor_conditions.torque  = Q
    motor_conditions.omega   = omega 

    return


# ----------------------------------------------------------------------------------------------------------------------
# compute_I_from_omega_and_V
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_current_from_RPM_and_voltage(motor, motor_conditions,conditions):
    """Calculates the motor's current from its RPM and voltage. 
    The following perperties of the motor are computed   
    motor_conditions.current     (numpy.ndarray): current      [A]
    motor_conditions.efficiency  (numpy.ndarray): efficiency   [unitless] 
      
    Assumptions:
        None
        
    Source:
        None

    Args: 
        gear_ratio                  (float): gear ratio                [unitless]
        speed_constant              (float): motor speed constant      [radian/s/V]
        resistance                  (float): motor internal resistnace [ohm]
        outputs.omega       (numpy.ndarray): angular velocity          [radian/s]
        gearbox_efficiency          (float): gearbox efficiency        [unitless]
        expected_current            (float): current                   [A]
        no_load_current             (float): no-load current           [A]
        inputs.volage       (numpy.ndarray): operating voltage         [V] 
        rotor_radius                (float): rotor radius              [m]

    Returns:
        None 
    """                      
    
    # Unpack
    KV    = motor.speed_constant
    Res   = motor.resistance
    omega = motor_conditions.omega
    V     = motor_conditions.voltage
    G     = motor.gear_ratio
    eta_G = motor.gearbox_efficiency
    exp_I = motor.expected_current
    I0    = motor.no_load_current + exp_I*(1-eta_G)
    
    # compute current 
    I    = (V-(omega*G)/KV)/Res 

    # Pack results 
    motor_conditions.current    = I
    motor_conditions.efficiency = (1-I0/I)*(1-I*Res/V)
    return



# ----------------------------------------------------------------------------------------------------------------------
# compute_V_and_I_from_omega_and_Kv
# ----------------------------------------------------------------------------------------------------------------------      
def compute_voltage_and_current_from_RPM_and_speed_constant(motor, motor_conditions,conditions):
    """Calculates the motor's voltage and current from its RPM and speed constant
    The following perperties of the motor are computed    
        motor_conditions.current     (numpy.ndarray): current    [A] 
        motor_conditions.volage      (numpy.ndarray): volage     [V] 
        motor_conditions.efficiency  (numpy.ndarray): efficiency [unitless]
        
    Assumptions:
        None 

    Source:
        None

    Args:
        gear_ratio                  (float): gear ratio                [unitless]
        speed_constant              (float): motor speed constant      [radian/s/V]
        resistance                  (float): motor internal resistnace [ohm]
        inputs.omega        (numpy.ndarray): angular velocity          [radian/s]
        gearbox_efficiency          (float): gearbox efficiency        [unitless]
        expected_current            (float): current                   [A]
        no_load_current             (float): no-load current           [A]
        inputs.volage       (numpy.ndarray): operating voltage         [V] 
        rotor_radius                (float): rotor radius              [m]

    Returns:
       None  
    """                      
           
    G      = motor.gear_ratio
    KV     = motor.speed_constant/G
    Q      = motor_conditions.torque
    omega  = motor_conditions.omega   
    Res    = motor.resistance     
    eta_G  = motor.gearbox_efficiency
    exp_I  = motor.expected_current
    I0     = motor.no_load_current + exp_I*(1-eta_G)
     
    V = (Q*KV+I0)*Res + omega/KV
    I = (V-omega/KV)/Res
    
    # Store results 
    motor_conditions.current    = I
    motor_conditions.efficiency = (1-I0/I)*(1-I*Res/V)
    motor_conditions.voltage    = V
    
    return


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core     import Data
from RCAIDE.Framework.Analyses import Analysis  

# ----------------------------------------------------------------------
#  Analysis
# ---------------------------------------------------------------------- 
class Weights(Analysis):
    """ This is a class that call the functions that computes the weight of 
    an aircraft depending on its configration
    
    Assumptions:
        None

    Source:
        N/A

    Inputs:
        None
        
    Outputs:
        None

    Properties Used:
         N/A
    """
    def __defaults__(self):
        """This sets the default values and methods for the weights analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """           
        self.tag      = 'weights' 
        self.method   = 'RCAIDE'

        self.vehicle  = None
        
        self.settings = Data()
        self.settings.use_max_fuel_weight                =  True
        self.settings.weight_reduction_factors           = Data()  
        self.settings.weight_reduction_factors.main_wing = 0.  # Reduction factors are proportional (.1 is a 10% weight reduction)
        self.settings.weight_reduction_factors.fuselage  = 0.  # Reduction factors are proportional (.1 is a 10% weight reduction)
        self.settings.weight_reduction_factors.empennage = 0.  # applied to horizontal and vertical stabilizers
        
        # FLOPS settings
        self.settings.FLOPS = Data() 
        self.settings.FLOPS.aeroelastic_tailoring_factor = 0.   # Aeroelastic tailoring factor [0 no aeroelastic tailoring, 1 maximum aeroelastic tailoring] 
        self.settings.FLOPS.strut_braced_wing_factor     = 0.   # Wing strut bracing factor [0 for no struts, 1 for struts]
        self.settings.FLOPS.composite_utilization_factor = 0.5  # Composite utilization factor [0 no composite, 1 full composite]
        
        # Raymer settings
        self.settings.Raymer = Data()
        self.settings.Raymer.fuselage_mounted_landing_gear_factor = 1. # 1. if false, 1.12 if true
                       
        
    def evaluate(self):
        """Evaluate the weight analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results 
        """
        # unpack
        vehicle = self.vehicle 
        results = RCAIDE.Library.Methods.Weights.Correlation_Buildups.Common.compute_operating_empty_weight(vehicle, settings=self.settings)

        # storing weigth breakdown into vehicle
        vehicle.weight_breakdown = results

        # updating empty weight
        vehicle.mass_properties.operating_empty = results.empty.total

        # done!
        return results        


# RCAIDE/Library/Attributes/Materials/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Solid                       import Solid
from .Unidirectional_Carbon_Fiber import Unidirectional_Carbon_Fiber
from .Bidirectional_Carbon_Fiber  import Bidirectional_Carbon_Fiber
from .Carbon_Fiber_Honeycomb      import Carbon_Fiber_Honeycomb
from .Epoxy                       import Epoxy
from .Aluminum_Rib                import Aluminum_Rib
from .Paint                       import Paint
from .Aluminum                    import Aluminum
from .Polyetherimide              import Polyetherimide
from .Acrylic                     import Acrylic
from .Steel                       import Steel
from .Nickel                      import Nickel
from .Titanium                    import Titanium
from .Magnesium                   import Magnesium


# RCAIDE/Library/Compoments/Wings/Vertical_Tail_All_Moving.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Vertical_Tail      import Vertical_Tail
from .All_Moving_Surface import All_Moving_Surface
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_wing_moment_of_inertia import  compute_wing_moment_of_inertia

# ---------------------------------------------------------------------------------------------------------------------- 
#  Vertical_Tail_All_Moving
# ----------------------------------------------------------------------------------------------------------------------     
class Vertical_Tail_All_Moving(Vertical_Tail, All_Moving_Surface):
    """
    A class representing an all-moving vertical tail that provides directional control 
    and stability.

    Attributes
    ----------
    tag : str
        Unique identifier for the surface, defaults to 'vertical_tail_all_moving'
        
    sign_duplicate : float
        Sign convention for duplicate surface deflection, defaults to -1.0
        (opposite deflection for yaw control)

    Notes
    -----
    The all-moving vertical tail combines the functions of a vertical stabilizer and 
    rudder by pivoting as a complete surface. It inherits from both Vertical_Tail and 
    All_Moving_Surface to combine their functionality.

    See Also
    --------
    RCAIDE.Library.Components.Wings.Vertical_Tail
        Base class providing vertical tail functionality
    RCAIDE.Library.Components.Wings.All_Moving_Surface
        Base class providing pivoting surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder
        Alternative yaw control using conventional rudder
    """ 

    def __defaults__(self):
        """
        Sets default values for the all-moving vertical tail attributes.
        
        Notes
        -----
        See All_Moving_Surface.__defaults__ and Wing.__defaults__ for 
        additional inherited attributes.
        """
        self.tag            = 'vertical_tail_all_moving'
        self.sign_duplicate = -1.0   
    
    def make_x_z_reflection(self):
        """
        Creates a reflected copy of the vertical tail over the x-z plane.

        Returns
        -------
        Component
            Reflected vertical tail with appropriate deflection sign

        Notes
        -----
        * Used when vertical tail's symmetric attribute is True
        * Deflection is reflected according to sign_duplicate convention
        * Should be called after setting control surface deflections
        """       
        wing                  = super().make_x_z_reflection()
        wing.deflection      *= -1*self.sign_duplicate
        wing.hinge_vector[1] *= -1
        return wing
    
    def moment_of_inertia(wing, center_of_gravity):
        """
        Computes the moment of inertia tensor for the vertical tail.

        Parameters
        ----------
        wing : Component
            Wing component data
        center_of_gravity : list
            Reference point coordinates for moment calculation

        Returns
        -------
        ndarray
            3x3 moment of inertia tensor
        """
        I = compute_wing_moment_of_inertia(wing, center_of_gravity) 
        return I  
        



# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/translate_conditions_to_dfdc_cases.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import  RCAIDE
from RCAIDE.Framework.Core import Data ,Units

# ---------------------------------------------------------------------------------------------------------------------- 
# Translate Conditions to DFDC Cases 
# ----------------------------------------------------------------------------------------------------------------------    
def translate_conditions_to_dfdc_cases(dfdc_analysis):
    """ Translate fligth conditions to DFDC Cases 

    Assumptions:
        N.A.

    Source:
        N.A.
    
    Inputs:
        dfdc_analysis (dict): DFDC analysis data structure  

    Outputs:
        None
    """
    # set up aerodynamic Conditions object
    template   = dfdc_analysis.settings.filenames.results_template 
    mach       = dfdc_analysis.training.mach     
    altitude   = dfdc_analysis.training.altitude        
    tip_mach   = dfdc_analysis.training.tip_mach
    ducted_fan = dfdc_analysis.geometry
    
    # first case is the design case 
    case            = Data() 
    atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data       = atmosphere.compute_values(ducted_fan.cruise.design_altitude)  
    rpm             = ducted_fan.cruise.design_angular_velocity/Units.rpm 
    string          = template.format(ducted_fan.cruise.design_freestream_velocity,rpm,ducted_fan.cruise.design_altitude)   
    case.tag        = string.replace(".", "_") + '.txt'
    case.velocity   = ducted_fan.cruise.design_freestream_mach * atmo_data.speed_of_sound[0,0]
    case.RPM        = rpm
    case.altitude   = ducted_fan.cruise.design_altitude / 1000
    dfdc_analysis.append_case(case)
    
    
    for i in range(len(mach)): 
        for j in range(len(tip_mach)):   
            for k in range(len(altitude)):     
                case            = Data() 
                atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
                atmo_data       = atmosphere.compute_values(altitude[k]) 
                a               = atmo_data.speed_of_sound[0,0]
                velocity        = mach[i] * a
                rpm             = ((tip_mach[j]*a) /ducted_fan.tip_radius)/Units.rpm
                string          = template.format(velocity,rpm,altitude[k])  
                case.tag        = string.replace(".", "_") + '.txt'
                case.velocity   = velocity
                case.RPM        = rpm
                case.altitude   = altitude[k]
                dfdc_analysis.append_case(case) 
    return 


# RCAIDE/Framework/Analyses/Mission/Segments/Single_Point/Set_Speed_Set_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Core                           import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate      import Evaluate
from RCAIDE.Library.Mission                          import Common,Segments
from RCAIDE.Library.Methods.skip                     import skip 

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Set_Speed_Set_Altitude
# ---------------------------------------------------------------------------------------------------------------------- 
class Set_Speed_Set_Altitude(Evaluate):
    """ This is a segment that is solved using a single point. A snapshot in time.
        We fix the speed and altitude. Throttle is solved from those.
    
        Assumptions:
        None
        
        Source:
        None
    """        
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # --------------------------------------------------------------------------------------------------------------
        #   User Inputs
        # --------------------------------------------------------------------------------------------------------------
        self.altitude                                = None
        self.air_speed                               = 10. * Units['km/hr']
        self.distance                                = 10. * Units.km
        self.linear_acceleration_x                   = 0.
        self.linear_acceleration_y                   = 0.  
        self.linear_acceleration_z                   = 0. # note that down is positive
        self.roll_rate                               = 0.
        self.pitch_rate                              = 0.  
        self.yaw_rate                                = 0.  
        self.state.numerics.number_of_control_points = 1   
         
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------     
        initialize                               = self.process.initialize 
        initialize.expand_state                  = skip
        initialize.differentials                 = skip
        initialize.conditions                    = Segments.Single_Point.Set_Speed_Set_Altitude.initialize_conditions 
        iterate                                  = self.process.iterate 
        iterate.initials.energy                  = skip
        iterate.unknowns.controls                = Common.Unpack_Unknowns.control_surfaces
        iterate.unknowns.mission                 = Common.Unpack_Unknowns.orientation  
        iterate.conditions.planet_position       = skip    
        iterate.conditions.acceleration          = skip
        iterate.conditions.angular_acceleration  = skip 
        iterate.conditions.weights               = skip
        iterate.residuals.flight_dynamics        = Common.Residuals.flight_dynamics
        post_process                             = self.process.post_process 
        post_process.inertial_position           = skip   
        
                
                
        return




# RCAIDE/Library/Missions/Common/Update/planet_position.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core  import Units 

# package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Update Planet Position
# ----------------------------------------------------------------------------------------------------------------------
def planet_position(segment):
    """ Updates the location of the vehicle relative to the Planet throughout the mission
    
        Assumptions:
        This is valid for small movements and times as it does not account for the rotation of the Planet beneath the vehicle
        
        Inputs:
        segment.state.conditions:
            freestream.velocity                      [meters/second]
            freestream.altitude                      [meters]
            frames.body.inertial_rotations           [Radians]
        segment.analyses.planet.mean_radius [meters]
        segment.state.numerics.time.integrate        [float]
            
        Outputs:
            segment.state.conditions:           
                frames.planet.latitude  [Radians]
                frames.planet.longitude [Radians]

        Properties Used:
        N/A
                                
    """        
    
    # unpack
    conditions = segment.state.conditions
    
    # unpack orientations and velocities
    V          = conditions.freestream.velocity[:,0]
    altitude   = conditions.freestream.altitude[:,0] 
    theta      = conditions.frames.body.inertial_rotations[:,1]
    psi        = segment.true_course        
    alpha      = conditions.aerodynamics.angles.alpha[:,0]
    I          = segment.state.numerics.time.integrate
    Re         = segment.analyses.planet.mean_radius  
         
    # The flight path and radius
    gamma     = theta - alpha
    R         = altitude + Re

    # Find the velocities and integrate the positions
    lamdadot  = (V/R)*np.cos(gamma)*np.cos(psi)
    lamda     = np.dot(I,lamdadot) / Units.deg # Latitude
    mudot     = (V/R)*np.cos(gamma)*np.sin(psi)/np.cos(lamda)
    mu        = np.dot(I,mudot) / Units.deg # Longitude

    # Reshape the size of the vectorss
    shape     = np.shape(conditions.freestream.velocity)
    mu        = np.reshape(mu,shape)
    lamda     = np.reshape(lamda,shape)
    phi       = np.array([[np.cos(psi),-np.sin(psi),0],[np.sin(psi),np.cos(psi),0],[0,0,1]])

    # Pack 
    lat                                           = conditions.frames.planet.latitude[0,0]
    lon                                           = conditions.frames.planet.longitude[0,0]
    conditions.frames.planet.latitude             = lat + lamda
    conditions.frames.planet.longitude            = lon + mu 
    conditions.frames.planet.true_course          = np.tile(phi[None,:,:],(len(V),1,1))    

    return 


# RCAIDE/Methods/Geometry/Two_Dimensional/Planform/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .convert_sweep                              import convert_sweep
from .fuselage_planform                          import fuselage_planform
from .horizontal_tail_planform                   import horizontal_tail_planform
from .vertical_tail_planform                     import vertical_tail_planform
from .wing_planform                              import wing_planform
from .horizontal_tail_planform_raymer            import horizontal_tail_planform_raymer 
from .wing_segmented_planform                    import wing_segmented_planform,  segment_properties 
from .vertical_tail_planform_raymer              import vertical_tail_planform_raymer
from .wing_fuel_volume                           import wing_fuel_volume
from .populate_control_sections                  import populate_control_sections
from .compute_span_location_from_chord_length    import compute_span_location_from_chord_length
from .compute_chord_length_from_span_location    import compute_chord_length_from_span_location 



# RCAIDE/Library/Methods/Aerodynamics/Common/Lift/fuselage_correction.py
# (c) Copyright 2023 Aerospace Research Community LLC
#  
# Created: Mar 2024 M. Carke     

# ----------------------------------------------------------------------------------------------------------------------
#  Fuselage Correction
# ----------------------------------------------------------------------------------------------------------------------
def fuselage_correction(state,settings,geometry):  
    """Corrects aircraft lift based on fuselage effects

    Assumptions:
        None

    Source:
        adg.stanford.edu (Stanford AA241 A/B Course Notes)

    Args:
        settings.fuselage_lift_correction (float): fuselage lift correction [unitless]
        state.conditions.
          freestream.mach_number          (numpy.ndarray): mach number      [unitless]
          aerodynamics.angles.alpha       (numpy.ndarray): angle of attack  [radians]
          aerodynamics.coefficients.lift  (numpy.ndarray): lift coefficient [unitless]

    Returns:
        aircraft_total_lift               (numpy.ndarray): lift coefficient [unitless]   
    """        
    # unpack 
    fus_correction  = settings.fuselage_lift_correction
    wings_lift_comp = state.conditions.aerodynamics.coefficients.lift.total
    
    # total lift, assuming one fuselage
    aircraft_total_lift = wings_lift_comp * fus_correction  

    state.conditions.aerodynamics.coefficients.lift.total = aircraft_total_lift

    return 


# RCAIDE/Framework/Analyses/Mission/Segments/Descent/Constant_Speed_Constant_Angle_Noise.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core                                 import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate   import Evaluate 
from RCAIDE.Library.Mission                      import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------  
#  Constant_Speed_Constant_Angle_Noise
# ----------------------------------------------------------------------------------------------------------------------   
class Constant_Speed_Constant_Angle_Noise(Evaluate):
    """ Fixed at a true airspeed the vehicle will descend at a constant angle. This is a specific segment for Noise.
        A vehicle performs a descent to landing in accordance with a certification points for landing noise.
        Don't use this segment unless you're planning on post process for noise. It is slower than the normal constant
        speed constand angle segment.
    
        Assumptions:
        The points are linearly spaced rather than chebyshev spaced to give the proper "microphone" locations.
        
        Source:
        None
    """     
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """            
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude_start    = None # Optional
        self.altitude_end      = 0.0 * Units.km
        self.descent_angle     = 3.  * Units.deg
        self.air_speed         = 100 * Units.m / Units.s
        self.true_course       = 0.0 * Units.degrees 
        
        self.state.numerics.discretization_method = RCAIDE.Library.Methods.Utilities.Chebyshev.linear_data 
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------  
        initialize                         = self.process.initialize  
        initialize.differentials_altitude  = Common.Initialize.differentials_altitude
        initialize.expand_state            = Segments.Descent.Constant_Speed_Constant_Angle_Noise.expand_state
        initialize.conditions              = Segments.Descent.Constant_Speed_Constant_Angle_Noise.initialize_conditions
        iterate                            = self.process.iterate   
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation          
        
        return




# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/TBL_TE_broadband_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Python Package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# Compute TBL-TE Broadband Noise 
# ---------------------------------------------------------------------------------------------------------------------- 
def TBL_TE_broadband_noise(f,r_e,L,U,M,R_c,Dbar_h,Dbar_l,R_delta_star_p,delta_star_p,delta_star_s,alpha_star):
    '''This computes the turbument boundary layer- trailing edge noise compoment of broadband noise using the method outlined by the 
    Brooks, Pope and Marcolini (BPM) Model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs:   
       R_delta_star_p - Reynolds number correlated to boundary layer thicness [m]
       delta_star_p   - boundary layer thickness of pressure side             [m]
       delta_star_s   - boundary layer thickness of suction side              [m]
       R_c            - Reynolds number                                       [-]
       alpha_star     - angle of attack blade section                         [deg]
       delta_p        - boundary layer of pressure section                    [m
       L              - length of blade section                               [m]
       U              - velocity of blade section                             [m/s]
       M              - Mach number                                           [-]
       c              - airfoil section chord                                 [m] 
       f              - frequency spectrum                                    [Hz]
       Dbar_h         - high frequency directivity term                       [-]
       Dbar_l         - low frequency directivity term                        [-]
       r_e            - distance from noise source to observer                [m] 
    
    
    Outputs 
       SPL_TBL_TE   - Sound pressure level of turbument boundary layer        [dB]
       
    Properties Used:
        N/A   
    '''      
     
    # Strouhal number
    St_p                  = f*delta_star_p/U    # eqn 31
    St_s                  = f*delta_star_s/U    # eqn 31
    St_1                  = 0.02*(M**(-0.6))      # eqn 32
    St_2                  = St_1*(10**(0.0054*((alpha_star - 1.33)**2)))
    St_2[alpha_star<1.33] = St_1[alpha_star<1.33]*1
    St_2[alpha_star>12.5] = St_1[alpha_star>12.5]*4.72 
    St_bar_1              = (St_1 + St_2)/2     # eqn 33 
    
    # Spectral Shape Functions
    St_peak   = St_bar_1 # can be St_1,St_2 or St_bar_1 
    A_p       = spectral_shape_function_A(St_p,St_peak,R_c)    # eqns 35 - 40
    A_s       = spectral_shape_function_A(St_s,St_peak,R_c)    # eqns 35 - 40
    A_prime   = spectral_shape_function_A(St_s,St_peak,R_c*3)  # eqns 35 - 40
    B         = spectral_shape_function_B(St_s,St_2,R_c)       # eqns 31 - 46    
    
    # Amplitude function 
    K_1                                  = amplitude_function_K_1(R_c)                             # eqn 47
    delta_K_1                            = amplitude_function_delta_K_1(alpha_star,R_delta_star_p) # eqn 48 
    K_2                                  = amplitude_function_K_2(alpha_star,M,K_1)                # eqn 49   
    SPL_alpha                            = 10*np.log10((delta_star_s*(M**5)*L*Dbar_h)/(r_e**2))  + B   + K_2                    # eqn 27 
    SPL_s                                = 10*np.log10((delta_star_s*(M**5)*L*Dbar_h)/(r_e**2))  + A_s + (K_1 - 3)              # eqn 26
    SPL_p                                = 10*np.log10((delta_star_p*(M**5)*L*Dbar_h)/(r_e**2))  + A_p + (K_1 - 3) + delta_K_1  # eqn 25   
   
    alpha_star_0_bool                    = np.zeros_like(alpha_star , dtype=bool) # 
    K_2_peak                             = np.max(K_2)
    alpha_switch_1                       = np.where(K_2 == K_2_peak)[0]
    alpha_switch_2                       = np.where(alpha_star>12.5)[0]
    alpha_star_0_bool[alpha_switch_1]    = True 
    alpha_star_0_bool[alpha_switch_2]    = True  
    
    SPL_p[alpha_star_0_bool]             = -np.inf # eqn 28
    SPL_s[alpha_star_0_bool]             = -np.inf # eqn 29 
    SPL_alpha[alpha_star_0_bool]         = 10*np.log10((delta_star_s[alpha_star_0_bool]*(M[alpha_star_0_bool]**5)*L[alpha_star_0_bool]*Dbar_l[alpha_star_0_bool])/(r_e[alpha_star_0_bool]**2))  + A_prime[alpha_star_0_bool]+ K_2[alpha_star_0_bool]  
             
    SPL_TBL_TE                           = 10*np.log10( 10**(SPL_alpha/10) + 10**(SPL_s/10) + 10**(SPL_p/10) ) # eqn 24 

    return  SPL_TBL_TE

def spectral_shape_function_A(St,St_peak,R_c):  
    a                  = abs(np.log10(St/St_peak))    # 37  
    a_0                = (-9.57E-13)*((R_c - 8.57E5)**2) + 1.13    # 38 
    a_0[R_c < 9.52E4]  =  0.57                            # 38 
    a_0[R_c > 8.57E5]  =  1.13                            # 38  
    
    A_min     = A_min_function(a)                          # eqn 35 
    A_min_a_0 = A_min_function(a_0)                        # eqn 35 
    A_max     = A_max_function(a)                          # eqn 36 
    A_max_a_0 = A_max_function(a_0)                        # eqn 36 
 
    A_R_a_0   = (-20 - A_min_a_0)/(A_max_a_0 - A_min_a_0)  # eqn 39 
    A         = A_min + A_R_a_0*abs(A_max -A_min )         # eqn 40
    
    return A 

def spectral_shape_function_B(St_s,St_2,R_c):
    '''
    COMMENTS NIRANJAN 
     
    '''
    b                   = abs(np.log10(St_s/St_2))                    # 43  
    b_0                 = (-4.48*(10**(-13)))*((R_c - 8.57E5)**2) + 0.56      # 44
    b_0[R_c < 9.52E4]   = 0.30                         # 44
    b_0[R_c > 8.57E5]   = 0.56                         # 44 
    
    B_min     = B_min_function(b)    # eqn 41
    B_min_b_0 = B_min_function(b_0)  # eqn 41
    B_max     = B_max_function(b)    # eqn 42 
    B_max_b_0 = B_max_function(b_0)  # eqn 42 
 
    B_R_b_0   = (-20 - B_min_b_0)/(B_max_b_0 - B_min_b_0)  # eqn 45
    B         = B_min + B_R_b_0*(B_max -B_min )         # eqn 40  
    return B

def A_min_function(a):   
    '''
    COMMENTS NIRANJAN 
     
    '''
    A_min          = -32.665*a +3.981 # eqn 35 
    A_min[a<0.204] = np.sqrt(67.552-886.788*(a[a<0.204]**2)) - 8.219 # eqn 35 
    A_min[a>0.244] = -142.795*a[a>0.244]**3 +  103.656*a[a>0.244]**2 - 57.757*a[a>0.244] + 6.006 # eqn 35  
    return A_min 

def A_max_function(a):    
    '''
    COMMENTS NIRANJAN 
     
    '''
    A_max          = -15.901*a + 1.098 # eqn 36 
    A_max[a<0.13]  = np.sqrt(67.552-886.788*(a[a<0.13]**2)) - 8.219 # eqn 36 
    A_max[a>0.321] = -4.669*a[a>0.321]**3 + 3.491*a[a>0.321]**2  - 16.699*a[a>0.321] + 1.149 # eqn 36  
    return A_max 


def B_min_function(b):   
    '''
    COMMENTS NIRANJAN 
     
    '''
    B_min          = -83.607*b + 8.138  # eqn 41
    B_min[b<0.13]  = np.sqrt(16.888-886.788*(b[b<0.13]**2)) - 4.109
    B_min[b>0.145] = -817.810*b[b>0.145]**3 +  355.210*b[b>0.145]**2 - 135.024*b[b>0.145] + 10.619 # eqn 41    
    return B_min 

def B_max_function(b):  
    '''
    COMMENTS NIRANJAN 
     
    '''
    B_max          = -31*33*b + 1.854  # eqn 42
    B_max[b<0.10]  = np.sqrt(16.888-886.788*(b[b<0.10]**2)) - 4.109 # eqn 42
    B_max[b>0.187] = -80.541*b[b>0.187]**3 +  44.174*b[b>0.187]**2 - 39.381*b[b>0.187] + 2.344 # eqn 42  
    return B_max 

def amplitude_function_K_1(R_c): 
    '''
    COMMENTS NIRANJAN 
     
    '''
    K_1             = - 9.0*np.log10(R_c) + 181.6   # eqn 47 
    K_1[R_c<2.47E5] = -4.31*np.log10(R_c[R_c<2.47E5]) + 156.3  # eqn 47
    K_1[R_c>8.0E5]  = 128.5   # eqn 47 
    return K_1

def amplitude_function_delta_K_1(alpha_star,R_delta_star_p): 
    '''
    COMMENTS NIRANJAN 
     
    ''' 
    delta_K_1                        = alpha_star*(1.43*np.log10(R_delta_star_p) - 5.29)   # eqn 48 
    delta_K_1[R_delta_star_p>5000] = 0                                                       # eqn 48     
    return delta_K_1

def amplitude_function_K_2(alpha_star,M,K_1):
    '''
    COMMENTS NIRANJAN 
     
    '''    
    gamma   = 27.094*M + 3.31  # eqn 50  
    gamma_0 = 23.43*M + 4.651  # eqn 50  
    beta    = 72.65*M + 10.74  # eqn 50
    beta_0  = -34.19*M - 13.82 # eqn 50   
    K_2                               = K_1 + np.sqrt(beta**2 - ((beta/gamma)**2)*((alpha_star-gamma_0)**2)) + beta_0      # eqn 49 
    K_2[alpha_star< (gamma_0-gamma)]  = K_1[alpha_star< (gamma_0-gamma)] -1000  # eqn 49 
    K_2[alpha_star> (gamma_0+gamma)]  = K_1[alpha_star> (gamma_0+gamma)] -12 # eqn 49 
    return K_2




# RCAIDE/Library/Methods/Propulsors/Converters/Compression_Nozzle/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
         
from .compute_compression_nozzle_performance import compute_compression_nozzle_performance
from .append_compression_nozzle_conditions   import append_compression_nozzle_conditions 


# RCAIDE/Library/Components/Thermal_Management/Batteries/Liquid_Cooled_Wavy_Channel.py
# 
# 
# Created:  Apr 2024 S. Shekar 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core                                                          import Units
from RCAIDE.Library.Attributes.Coolants.Glycol_Water                                import Glycol_Water
from RCAIDE.Library.Components.Component                                            import Component  
from RCAIDE.Library.Attributes.Materials.Aluminum                                   import Aluminum
from RCAIDE.Library.Components                                                      import Component
from RCAIDE.Library.Components.Component                                            import Container
from RCAIDE.Library.Methods.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel import  wavy_channel_rating_model,append_wavy_channel_conditions,append_wavy_channel_segment_conditions 
from RCAIDE.Library.Plots.Thermal_Management                                        import plot_wavy_channel_conditions
# ----------------------------------------------------------------------------------------------------------------------
# Liquid_Cooled_Wavy_Channel_Heat_Acquisition_System
# ----------------------------------------------------------------------------------------------------------------------
class Liquid_Cooled_Wavy_Channel(Component):
    '''
       Wavy Channel Heat Acqusition System
    '''
    
    def __defaults__(self):  
        """This sets the default values.
        
        Assumptions:
           The wavy channel heat Acquisition loops through the battery pack.
           The coolant is assumed to be Glycol Water unless specified otherwise. 
           The geometry parameters are set based on nomrinal values to be further optmized.
           
        Source:
           None
        
        """            
         
        self.tag                           = 'wavy_channel_heat_acquisition' 
        self.heat_transfer_efficiency      = 1
        self.coolant                       = Glycol_Water()
        self.coolant_Reynolds_number       = 1.
        self.coolant_velocity              = 1.
        self.coolant_flow_rate             = 1
        self.coolant_inlet_temperature     = None
        self.coolant_hydraulic_diameter    = 1.
        self.channel_side_thickness        = 0.001                  # Thickness of the Chanel through which condcution occurs 
        self.channel_top_thickness         = 0.001                  # Thickness of Channel on the top where no conduction occurs
        self.channel_width                 = 0.005                  # width of the channel 
        self.channel_height                = 0.003                  # height of the channel 
        self.channel_contact_angle         = 47.5 * Units.degrees   # Contact Arc angle in degrees    
        self.channel                       = Aluminum()
        self.channel_aspect_ratio          = 1. 
        self.channels_per_module           = 1
        self.battery_contact_area          = 1.
        self.contact_area_per_module       = 1.  
        self.power_draw                    = 1. 
        self.single_side_contact           = True 
        self.design_heat_removed           = None   
        self.percent_operation             = 1.0
        self.type                          = 'Liquid'
        
        return
    
    def __init__ (self, distributor=None):
        
        """This creates Reservoir and Heat Exchanger containers if it does not exist on
            the coolant line as a liquid cooled system requires these and cannot operate without.
    
        Assumptions:
            None
        
        Source:
            None
        """                
        
    def append_operating_conditions(self,segment,coolant_line):
        append_wavy_channel_conditions(self,segment,coolant_line)
        return
    
    def append_segment_conditions(self, segment,coolant_line, conditions):
        append_wavy_channel_segment_conditions(self, segment,coolant_line, conditions)
        return
    
    def compute_thermal_performance(self,battery,bus,coolant_line,Q_heat_gen,T_cell,state,delta_t,t_idx):
        T_battery_current =  wavy_channel_rating_model(self, battery,bus,coolant_line, Q_heat_gen, T_cell, state, delta_t, t_idx)
        return  T_battery_current
    
    def plot_operating_conditions(self, results, coolant_line,save_filename, save_figure,show_legend,file_type , width, height):
        plot_wavy_channel_conditions(self, results, coolant_line,save_filename,save_figure,show_legend,file_type , width, height)
        return


# RCAIDE/Methods/Noise/Common/convert_to_third_octave_band.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Convert to Third Octave Band 
# ----------------------------------------------------------------------------------------------------------------------       
def convert_to_third_octave_band(SPL,f,settings): 
    """This method converts the SPL spectrum from blade harmonic passing frequency
    to thrid octave spectrum
    
    Assumptions:
        N/A

    Source: 

    Inputs:
        SPL                    - sound pressure level                          [dB] 
        f                      - blade passing spectrum frequencies            [Hz]
        settings.    
            center_frequencies - center frequencies of the 1/3 octave spectrum [dB]  
            lower_frequencies  - lower frequencies of the 1/3 octave spectrum  [dB]
            upper_frequencies  - upper frequencies of the 1/3 octave spectrum  [dB]
        

    Outputs:
        SPL_third_octave       - SPL in the 1/3 octave spectrum                [dB] 

    Properties Used:
        N/A 
        
    """  
    # unpack 
    cf               = settings.center_frequencies
    lf               = settings.lower_frequencies
    uf               = settings.upper_frequencies 
    dim_cpt          = len(SPL[:,0,0])
    dim_mic          = len(SPL[0,:,0]) 
    num_cf           = len(cf)
    num_f            = len(f[0,0,:]) 
    
    uf_vals          = np.tile(uf[None,None,:,None],(dim_cpt,dim_mic,1,num_f))
    lf_vals          = np.tile(lf[None,None,:,None],(dim_cpt,dim_mic,1,num_f))
    f_vals           = np.tile(f[:,:,None,:],(1,1,num_cf,1))
    SPL_vals         = np.tile(SPL[:,:,None,:],(1,1,num_cf,1))
     
    upper_bool       = (f_vals  <= uf_vals)
    lower_bool       = (lf_vals <= f_vals)
    boolean          = np.logical_and(upper_bool,lower_bool)
    SPL_array        = boolean*SPL_vals
    p_prefs          = 10**(SPL_array/10)
    SPL_third_octave = 10*np.log10(np.sum(boolean*p_prefs, axis =3)) 
    SPL_third_octave[np.isinf(SPL_third_octave)]  = 0 
    
    return SPL_third_octave



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(vehicle, wing):
    """ Calculates vertical tail weight based on Raymer method

        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                    [dimensionless]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
                -.flight_envelope.ultimate_load: ultimate load factor (default: 3.75)
                -.wings['main_wing']: data dictionary with properties of main wing
                    -.aerodynamic_center: aerodynamic center as measured from root leading edge
                    -.origin: root of main wing as measured from nose of aircraft
            wing    - data dictionary with specific tail properties              [dimensionless]
                -.areas.reference: tail surface area                            [m^2}
                -.origin: location of tail measured from nose
                -.aerodynamic_center: location of ac measured from leading edge
                -.sweeps.quarter_chord: quarter chord sweep of tail             [rad]
                -.thickness_to_chord: t/c of tail

        Outputs:
              tail_weight: vertical tail weight                                [kilograms]

        Properties Used:
            N/A
    """
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs
    t_tail_flag = wing.t_tail
    wing_origin = wing.origin[0][0] / Units.ft
    wing_ac     = wing.aerodynamic_center[0] / Units.ft
    main_origin = vehicle.wings['main_wing'].origin[0][0] / Units.ft
    main_ac     = vehicle.wings['main_wing'].aerodynamic_center[0] / Units.ft
    Svt         = wing.areas.reference / Units.ft ** 2
    sweep       = wing.sweeps.quarter_chord
    Av          = wing.aspect_ratio
    t_c         = wing.thickness_to_chord 
    Nult        = vehicle.flight_envelope.ultimate_load
    
    H = 0
    if t_tail_flag:
        H = 1
    Lt = (wing_origin + wing_ac - main_origin - main_ac)
    Kz = Lt
    tail_weight = 0.0026 * (1 + H) ** 0.225 * DG ** 0.556 * Nult ** 0.536 \
                  * Lt ** (-0.5) * Svt ** 0.5 * Kz ** 0.875 * np.cos(sweep) ** (-1) * Av ** 0.35 * t_c ** (-0.5)
    return tail_weight * Units.lbs



# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/compressibility_drag_total.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
   
from RCAIDE.Library.Components.Wings          import Main_Wing
from RCAIDE.Library.Methods.Utilities         import Cubic_Spline_Blender  
from .wave_drag                               import wave_drag
from .drag_divergence                         import drag_divergence 
from .supersonic_wave_drag_volume_raymer      import supersonic_wave_drag_volume_raymer
from .supersonic_wave_drag_volume_sears_haack import supersonic_wave_drag_volume_sears_haack

# package imports
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Compressibility Drag Total
# ----------------------------------------------------------------------------------------------------------------------  
def compressibility_drag(state,settings,geometry):
    """Computes compressibility drag for full aircraft including volume drag

    Assumptions:
    None

    Source:
    None

    Args:   
    settings.
      begin_drag_rise_mach_number                                    [Unitless]
      end_drag_rise_mach_number                                      [Unitless]
      peak_mach_number                                               [Unitless]
      transonic_drag_multiplier                                      [Unitless]
      volume_wave_drag_scaling                                       [Unitless]
    state.conditions.aerodynamics.lift_breakdown.compressible_wings  [Unitless]
    state.conditions.freestream.mach_number                          [Unitless]
    geometry.maximum_cross_sectional_area                            [m^2] (used in subfunctions)
    geometry.total_length                                            [m]   (used in subfunctions)
    geometry.reference_area                                          [m^2]
    geometry.wings                             

    Returns:
    total_compressibility_drag                                       [Unitless]

    Properties Used:
    None
    """     

    # Unpack
    conditions  = state.conditions
    Mach        = conditions.freestream.mach_number 
    Cl          = conditions.aerodynamics.coefficients.lift.total  
      
    # supersonic 
    low_mach_cutoff  = settings.supersonic.begin_drag_rise_mach_number
    high_mach_cutoff = settings.supersonic.end_drag_rise_mach_number
    peak_mach        = settings.supersonic.peak_mach_number
    peak_factor      = settings.supersonic.transonic_drag_multiplier
    scaling_factor   = settings.supersonic.volume_wave_drag_scaling
    
    if settings.supersonic.wave_drag_type == 'Raymer':
        wave_drag_volume = supersonic_wave_drag_volume_raymer
    elif settings.supersonic.wave_drag_type == 'Sears-Haack':
        wave_drag_volume = supersonic_wave_drag_volume_sears_haack
    else:
        raise NotImplementedError    
    
    if settings.supersonic.cross_sectional_area_calculation_type != 'Fixed':
        raise NotImplementedError

    low_cutoff_volume_total  = np.zeros_like(Mach)
    high_cutoff_volume_total = np.zeros_like(Mach)     
    low_cutoff_volume_total  = drag_divergence(low_mach_cutoff*np.ones_like(Mach), geometry,Cl) 
    high_cutoff_volume_total = wave_drag_volume(geometry,low_mach_cutoff*np.ones_like(Mach),scaling_factor)
    
    peak_volume_total = high_cutoff_volume_total*peak_factor
     
    # subsonic side
    a1 = (low_cutoff_volume_total-peak_volume_total)/(low_mach_cutoff-peak_mach)/(low_mach_cutoff-peak_mach)
    
    # supersonic side
    a2 = (high_cutoff_volume_total-peak_volume_total)/(high_mach_cutoff-peak_mach)/(high_mach_cutoff-peak_mach) 
    
    # Shorten cubic Hermite spline
    sub_spline = Cubic_Spline_Blender(low_mach_cutoff, peak_mach-(peak_mach-low_mach_cutoff)*3/4)
    sup_spline = Cubic_Spline_Blender(peak_mach,high_mach_cutoff)
    sub_h00    = lambda M:sub_spline.compute(M)
    sup_h00    = lambda M:sup_spline.compute(M) 
    
    low_inds = Mach[:,0]<peak_mach
    hi_inds  = Mach[:,0]>=peak_mach

    cd_c_v_base                  = np.zeros_like(Mach) 
    cd_c_v_base[low_inds]        = drag_divergence(Mach[low_inds], geometry,Cl[low_inds]) 
    cd_c_l_base                  = lift_wave_drag(conditions, settings, geometry)
    cd_c_v_base[Mach>=peak_mach] = wave_drag_volume(geometry, Mach[Mach>=peak_mach], scaling_factor) 
    
    cd_c_v = np.zeros_like(Mach)
    cd_c_v[low_inds] = cd_c_v_base[low_inds]*(sub_h00(Mach[low_inds])) + transonic_drag_function(Mach[low_inds],a1[low_inds], peak_mach, peak_volume_total[low_inds])*(1-sub_h00(Mach[low_inds]))
    cd_c_v[hi_inds]  = transonic_drag_function(Mach[hi_inds],a2[hi_inds], peak_mach, peak_volume_total[hi_inds])*(sup_h00(Mach[hi_inds])) + cd_c_v_base[hi_inds]*(1-sup_h00(Mach[hi_inds]))

    if peak_mach<1.01:
        print('Warning: a peak Mach number of less than 1.01 will cause a small discontinuity in lift wave drag')
    cd_c_l = cd_c_l_base*(1-sup_h00(Mach))
    
    cd_c = cd_c_v + cd_c_l 
    
    # Save drag breakdown 
    drag = conditions.aerodynamics.coefficients.drag.compressible.total   = cd_c
        
    return  

# ---------------------------------------------------------------------------------------------------------------------- 
#  transonic_drag_function
# ---------------------------------------------------------------------------------------------------------------------- 
def transonic_drag_function(Mach,a_vertex, peak_mach, peak_volume_total):
    """  parabolic approximation of drag rise in the transonic region
    
    Assumptions:
    Basic fit

    Source:
    None 

    Args:
    Mach
    a_vertex
    peak_mach
    peak_volume_total

    Returns:
    transonic_drag
    """ 
    transonic_drag = a_vertex*(Mach-peak_mach)*(Mach-peak_mach)+peak_volume_total
    transonic_drag = transonic_drag.reshape(np.shape(Mach))
    return transonic_drag

# ---------------------------------------------------------------------------------------------------------------------- 
# lift_wave_drag
# ---------------------------------------------------------------------------------------------------------------------- 
def lift_wave_drag(conditions,configuration,geometry):
    """Determine lift wave drag for supersonic speeds

    Assumptions:
    Basic fit

    Source:
    http://aerodesign.stanford.edu/aircraftdesign/aircraftdesign.html (Stanford AA241 A/B Course Notes)

    Args:
    conditions.freestream.mach_number [-]
    configuration                     (passed to another function)
    wing.areas.reference              [m^2]
    Sref_main                         [m^2] Main reference area

    Returns:
    cd_c_l                            [-] Wave drag CD due to lift 
    """ 
    for wing in  geometry.wings:
        if isinstance(wing, Main_Wing):  
            # Unpack Mach number
            Mach       = conditions.freestream.mach_number
        
            # Initalize cd arrays
            cd_c_l = np.zeros_like(Mach) 
        
            # Calculate wing values at all Mach numbers
            cd_lift_wave = wave_drag(conditions,wing)
        
            # Pack supersonic results into correct elements
            cd_c_l[Mach >= 1.01] = cd_lift_wave[0:len(Mach[Mach >= 1.01]),0] 

    return cd_c_l




# RCAIDE/Framework/Analyses/Vehicle.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from .Analysis import Analysis

# ----------------------------------------------------------------------------------------------------------------------
# Vehicle
# ----------------------------------------------------------------------------------------------------------------------  
class Vehicle(Analysis.Container):
    """ RCAIDE.Framework.Analyses.Vehicle()
    
        The Vehicle Analyses Container Class
        
            Assumptions:
            None
            
            Source:
            N/A
    """
    def __defaults__(self):
        """This sets the default analyses to be applied to the vehicle.
        
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                None
                
                Outputs:
                None
                
                Properties Used:
                N/A
        """
        self.aerodynamics = None
        self.atmosphere   = None
        self.costs        = None
        self.energy       = None
        self.noise        = None
        self.emissions    = None
        self.planet       = None
        self.sizing       = None
        self.stability    = None
        self.weights      = None

    def append(self,analysis):
        """This is used to add new analyses to the container.
        
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                Analysis to be added
                
                Outputs:
                None
                
                Properties Used:
                N/A
        """

        key = self.get_root(analysis) 
        self[key] = analysis 
    _analyses_map = None

    def __init__(self,*args,**kwarg):
        """This sets the initialization behavior of the vehicle analysis
           container. Maps analysis paths to string keys.
           
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                None
                
                Outputs:
                None
                
                Properties Used:
                N/A
        """

        Analysis.Container.__init__(self,*args,**kwarg)

        self._analyses_map = { 
            RCAIDE.Framework.Analyses.Weights.Weights           : 'weights'      ,
            RCAIDE.Framework.Analyses.Aerodynamics.Aerodynamics : 'aerodynamics' ,
            RCAIDE.Framework.Analyses.Stability.Stability       : 'stability'    ,
            RCAIDE.Framework.Analyses.Energy.Energy             : 'energy'       ,
            RCAIDE.Framework.Analyses.Atmospheric.Atmospheric   : 'atmosphere'   ,
            RCAIDE.Framework.Analyses.Planets.Planet            : 'planet'       ,
            RCAIDE.Framework.Analyses.Noise.Noise               : 'noise'        ,
            RCAIDE.Framework.Analyses.Emissions.Emissions       : 'emissions'    ,
        }

    def get_root(self,analysis):

        """ This is used to determine the root of the analysis path associated
            with a particular analysis key by the analysis map.
            
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                Analysis key to be checked
                
                Outputs:
                Path root of analysis
                
                Properties Used:
                N/A
                
                
        """
        for analysis_type, analysis_root in self._analyses_map.items():
            if isinstance(analysis,analysis_type):
                break
        else:
            raise Exception("Unable to place analysis type %s" % analysis.typestring())

        return analysis_root





# RCAIDE/Library/Components/Thermal_Management/Heat_Exchangers/__init__.py

"""
Module containing heat exchanger components for thermal management systems. This module 
provides classes for different heat exchanger configurations including cross-flow and 
cryogenic designs.

See Also
--------
RCAIDE.Library.Components.Thermal_Management.Batteries
    Related module for battery cooling applications
RCAIDE.Library.Components.Thermal_Management.Accessories
    Related module for pumps and fans used with heat exchangers
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .Cross_Flow_Heat_Exchanger  import Cross_Flow_Heat_Exchanger
from .Cryogenic_Heat_Exchanger   import Cryogenic_Heat_Exchanger



# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/append_cross_flow_heat_exchanger_conditions.py
# 
# Created:  Aug 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions


# ---------------------------------------------------------------------------------------------------------------------- 
#  append_cross_flow_heat_exchanger_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_cross_flow_heat_exchanger_conditions(cross_flow_hex,segment,coolant_line):
    """ Packs the initial cross flow heat exchanger conditions
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
             cross_flow_hex          (data structure)              [None]
           
        Outputs:
            segment
               coolant_mass_flow_rate                              [kg/s]
               power                                               [watts]
               inlet_air_temperature                               [Kelvin]
               outlet_coolant_temperature                          [Kelvin]
               air_mass_flow_rate                                  [kg/s]
               air_inlet_pressure                                  [Pascal]
               coolant_inlet_pressure                              [Pascal]
               pressure_diff_air                                   [Pascal]
               effectiveness_HEX                                   [unitless]
               
        Properties Used:
        None
    """        
    
    atmosphere    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    alt           = -segment.conditions.frames.inertial.position_vector[:,2] 
    if segment.temperature_deviation != None:
        temp_dev = segment.temperature_deviation    
    atmo_data    = atmosphere.compute_values(altitude = alt,temperature_deviation=temp_dev)
    
    ones_row                                                                                         = segment.state.ones_row
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag]                            = Conditions()
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].coolant_mass_flow_rate     = 0 * ones_row(1)  
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].power                      = 0 * ones_row(1)  
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].inlet_air_temperature      = atmo_data.temperature[0,0]* ones_row(1) 
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].outlet_coolant_temperature = atmo_data.temperature[0,0]* ones_row(1) 
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].air_mass_flow_rate         = 0 * ones_row(1) 
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].air_inlet_pressure         = 0 * ones_row(1) 
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].coolant_inlet_pressure     = 0 * ones_row(1) 
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].pressure_diff_air          = 0 * ones_row(1)
    segment.state.conditions.energy[coolant_line.tag][cross_flow_hex.tag].effectiveness_HEX          = 0 * ones_row(1)
    
    return

def append_cross_flow_hex_segment_conditions(cross_flow_hex,segment,bus,coolant_line,conditions):
    """Sets the initial cross flow heat exchanger properties at the start of each segment as the last point from the previous segment 
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
         cross_flow_hex            (data structure)              [None]
               
        Outputs:
        None
    
        Properties Used:
        None
    """    

    cross_flow_hex_conditions = conditions[coolant_line.tag][cross_flow_hex.tag]
    if segment.state.initials:  
        cross_flow_hex_initials                                   = segment.state.initials.conditions.energy[coolant_line.tag][cross_flow_hex.tag]
        cross_flow_hex_conditions.coolant_mass_flow_rate[:,0]     = cross_flow_hex_initials.coolant_mass_flow_rate[-1,0]     
        cross_flow_hex_conditions.power[:,0]                      = cross_flow_hex_initials.power[-1,0]     
        cross_flow_hex_conditions.air_inlet_pressure[:,0]         = cross_flow_hex_initials.air_inlet_pressure[-1,0]       
        cross_flow_hex_conditions.coolant_inlet_pressure[:,0]     = cross_flow_hex_initials.coolant_inlet_pressure[-1,0] 
        cross_flow_hex_conditions.inlet_air_temperature[:,0]      = cross_flow_hex_initials.inlet_air_temperature[-1,0]   
        cross_flow_hex_conditions.pressure_diff_air[:,0]          = cross_flow_hex_initials.pressure_diff_air[-1,0]                     
        cross_flow_hex_conditions.air_mass_flow_rate[:,0]         = cross_flow_hex_initials.air_mass_flow_rate[-1,0]     
        cross_flow_hex_conditions.outlet_coolant_temperature[:,0] = cross_flow_hex_initials.outlet_coolant_temperature[-1,0]  
        cross_flow_hex_conditions.effectiveness_HEX[:,0]          = cross_flow_hex_initials.effectiveness_HEX[-1,0]
        
        return
    


# RCAIDE/Framework/Analyses/Mission/Segments/Climb/Constant_Throttle_Constant_Speed.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core                       import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate  import Evaluate
from RCAIDE.Framework.Mission.Segments.Cruise    import Constant_Throttle_Constant_Altitude
from RCAIDE.Library.Mission                      import Common,Segments
from RCAIDE.Framework.Analyses                   import Process  

# ----------------------------------------------------------------------------------------------------------------------
# Constant_Throttle_Constant_Speed
# ---------------------------------------------------------------------------------------------------------------------- 
 
class Constant_Throttle_Constant_Speed(Evaluate):
    """ Climb at a constant throttle setting and true airspeed. This segment may not always converge as the vehicle 
        could be deficient in thrust. Useful as a check to see the climb rate at the top of climb.
    
        Assumptions:
        You set a reasonable throttle setting that can provide enough thrust.
        
        Source:
        None
    """     
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude_start    = None # Optional
        self.altitude_end      = 10. * Units.km
        self.throttle          = 0.5
        self.air_speed         = None
        self.true_course       = 0.0 * Units.degrees       

        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------      
        initialize                         = self.process.initialize   
        initialize.conditions              = Segments.Climb.Constant_Throttle_Constant_Speed.initialize_conditions 
        
        iterate                            = self.process.iterate 
         
        # Update Conditions
        iterate.conditions = Process()
        iterate.conditions.velocities                 = Segments.Climb.Constant_Throttle_Constant_Speed.update_velocity_vector_from_wind_angle 
        iterate.conditions.angles                     = Segments.Climb.Constant_Throttle_Constant_Speed.unpack_body_angle  
        iterate.conditions.differentials_altitude     = Segments.Climb.Constant_Throttle_Constant_Speed.update_differentials_altitude  
        iterate.conditions.differentials              = Common.Update.differentials_time 
        iterate.conditions.orientations               = Common.Update.orientations   
        iterate.conditions.acceleration               = Common.Update.acceleration          
        iterate.conditions.atmosphere                 = Common.Update.atmosphere
        iterate.conditions.gravity                    = Common.Update.gravity
        iterate.conditions.freestream                 = Common.Update.freestream 
        iterate.conditions.energy                     = Common.Update.thrust
        iterate.conditions.aerodynamics               = Common.Update.aerodynamics
        iterate.conditions.stability                  = Common.Update.stability
        iterate.conditions.weights                    = Common.Update.weights
        iterate.conditions.forces                     = Common.Update.forces
        iterate.conditions.moments                    = Common.Update.moments
        iterate.conditions.planet_position            = Common.Update.planet_position
        iterate.residuals.flight_dynamics             = Common.Residuals.flight_dynamics 
        
        return




# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/aero_coeff.py
# 
# 
# Created:  Dec 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Data

# pacakge imports  
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# aero_coeff
# ----------------------------------------------------------------------------------------------------------------------   
def aero_coeff(x,y,cp,al,npanel):
    """Compute airfoil force and moment coefficients about 
                    the quarter chord point          

    Assumptions:
    None

    Source:
    None                                                                    
                                                                   
    Inputs:                                                     
    x       -  Vector of x coordinates of the surface nodes      
    y       -  Vector of y coordinates of the surface nodes       
    cp      -  Vector of coefficients of pressure at the nodes  
    al      -  Angle of attack in radians                             
    npanel  -  Number of panels on the airfoil               
                                                                 
    Outputs:                                           
    cl      -  Airfoil lift coefficient                
    cd      -  Airfoil drag coefficient                                 
    cm      -  Airfoil moment coefficient about the c/4     
     
   Properties Used:
    N/A
    """    
     
    dx      = x[1:]-x[:-1]
    dy      = y[1:]-y[:-1]
    xa      = 0.5*(x[1:] +x[:-1])-0.25
    ya      = 0.5*(y[1:] +y[:-1]) 
    dcn     = -cp*dx  
    dca     = cp*dy   
    
    # compute differential forces
    cn      = np.sum(dcn,axis=0).T
    ca      = np.sum(dca,axis=0).T
    cm      = np.sum((-dcn*xa + dca*ya),axis=0).T
    
    # orient normal and axial forces 
    cl      = cn*np.cos(al) - ca*np.sin(al) 
    cdpi    = cn*np.sin(al) + ca*np.cos(al)  
    
    # pack results
    AERO_RES = Data(
        cl   = cl, 
        cdpi = cdpi,
        cm   = cm)
    
    return AERO_RES


# RCAIDE/Methods/Energy/Sources/Battery/Common/compute_net_generated_battery_heat.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# package imports 
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common 
def compute_net_generated_battery_heat(n_total,battery,Q_heat_gen,numerics):
    '''Computes the net heat generated in a battery during cycling. 

    Assumptions:
    1) Battery pack cell heat transfer can be modelled as a cooling columns in a cross-flow
    2) Isothermal battery cell - the temperature at the center of the cell is the same at 
    the surface of the cell

    Source: 
    Heat Transfer Model:
    Pakowski, Zdzisław. "Fundamentals of Heat and Mass Transfer, Frank P Incropera,
    David P DeWitt, Theodore L Bergman, Adrienne S Lavine, J. Wiley & Sons, Hoboken
    NJ (2007), 997 pp." (2007): 1683-1684.,  Chapter 7 pg 437-446 

    Inputs:  
        battery. 
              h                         (heat transfer coefficient)  [W/(m^2*K)] 
              As_cell                   (battery cell surface area)  [meters^2]
              H_cell                    (battery cell height)        [meters]
              T_ambient                 (ambient temperature)        [Kelvin]
              T_current                 (pack temperature)           [Kelvin]
              T_cell                    (battery cell temperature)   [Kelvin] 
              heat_transfer_efficiency                               [unitless]
      
      Outputs:
        battery. 
             net_power                                               [Watts] 
 

    Properties Used:
    None 
    '''

    h                        = battery.convective_heat_transfer_coefficient
    As_cell                  = battery.cell.surface_area 
    D_cell                   = battery.cell.diameter                     
    H_cell                   = battery.cell.height     
    T_ambient                = battery.ambient_temperature 
    T_current                = battery.temperature      
    T_cell                   = battery.cell.temperature       
    cell_mass                = battery.cell.mass    
    Cp                       = battery.cell.specific_heat_capacity       
    I                        = numerics.time.integrate      
    heat_transfer_efficiency = battery.heat_transfer_efficiency   

    # Calculate the current going into one cell   
    Nn                = battery.module.geometrtic_configuration.normal_count            
    Np                = battery.module.geometrtic_configuration.parallel_count   
    n_total           = Nn*Np       
    n_total_module    = Nn*Np    
    
    if n_total == 1: 
        # Using lumped model   
        Q_convec       = h*As_cell*(T_cell - T_ambient)
        P_net          = Q_heat_gen - Q_convec

    else: 
        # Chapter 7 pg 437-446 of Fundamentals of heat and mass transfer 
        S_T             = battery.module.geometrtic_configuration.normal_spacing          
        S_L             = battery.module.geometrtic_configuration.parallel_spacing   
        coolant         = battery.cooling_fluid
        K_coolant       = coolant.thermal_conductivity 
        Cp_coolant      = coolant.specific_heat_capacity  
        V_coolant       = coolant.cooling_flowspeed
        rho_coolant     = coolant.density 
        nu_coolant      = coolant.kinematic_viscosity     
        Pr_coolant      = coolant.prandtl_number  

        S_D = np.sqrt(S_T**2+S_L**2)
        if 2*(S_D-D_cell) < (S_T-D_cell):
            V_max = V_coolant*(S_T/(2*(S_D-D_cell)))
        else:
            V_max = V_coolant*(S_T/(S_T-D_cell))

        T        = (T_ambient+T_current)/2   
        Re_max   = V_max*D_cell/nu_coolant   
        if all(Re_max) > 10E2: 
            C = 0.35*((S_T/S_L)**0.2) 
            m = 0.6 
        else:
            C = 0.51
            m = 0.5  

        Pr_w_coolant          = coolant.compute_prandtl_number(T)            
        Nu                    = C*(Re_max**m)*(Pr_coolant**0.36)*((Pr_coolant/Pr_w_coolant)**0.25)           
        h                     = Nu*K_coolant/D_cell
        Tw_Ti                 = (T - T_ambient)
        Tw_To                 = Tw_Ti * np.exp((-np.pi*D_cell*n_total_module*h)/(rho_coolant*V_coolant*Nn*S_T*Cp_coolant))
        dT_lm                 = (Tw_Ti - Tw_To)/np.log(Tw_Ti/Tw_To)
        Q_convec              = heat_transfer_efficiency*h*np.pi*D_cell*H_cell*n_total_module*dT_lm 
        Q_convec[Tw_Ti == 0.] = 0.
        P_net                 = Q_heat_gen*n_total_module - Q_convec 

    dT_dt     = P_net/(cell_mass*n_total_module*Cp)
    T_current = T_current[0] + np.dot(I,dT_dt)  
    T_current[T_ambient>T_current] = T_ambient[T_ambient>T_current]
    
    return  T_current


# RCAIDE/Library/Missions/Segments/Common/Initialize/Frames.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Planet Position
# ----------------------------------------------------------------------------------------------------------------------
def planet_position(segment):
    """ Sets the initial location of the vehicle relative to the planet at the start of the segment
    
        Assumptions:
        Only used if there is an initial condition
        
        Inputs:
            segment.state.initials.conditions:
                frames.planet.longitude [Radians]
                frames.planet.latitude  [Radians]
            segment.longitude           [Radians]
            segment.latitude            [Radians]

        Outputs:
            segment.state.conditions:           
                frames.planet.latitude  [Radians]
                frames.planet.longitude [Radians]

        Properties Used:
        N/A
                                
    """        
    
    if segment.state.initials:
        longitude_initial = segment.state.initials.conditions.frames.planet.longitude[-1,0]
        latitude_initial  = segment.state.initials.conditions.frames.planet.latitude[-1,0] 
    elif 'latitude' in segment:
        longitude_initial = segment.longitude
        latitude_initial  = segment.latitude      
    else:
        longitude_initial = 0.0
        latitude_initial  = 0.0

    segment.state.conditions.frames.planet.longitude[:,0] = longitude_initial
    segment.state.conditions.frames.planet.latitude[:,0]  = latitude_initial    

    return 




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------
def wing_planform(wing):
    """Computes standard wing planform values.

    Assumptions:
    Trapezoidal wing with no leading/trailing edge extensions

    Source:
    None

    Inputs:
    wing.
      areas.reference          [m^2]
      taper                    [-]
      sweeps.quarter_chord     [radians]
      aspect_ratio             [-]
      thickness_to_chord       [-]
      dihedral                 [radians]
      vertical                 <boolean> Determines if wing is vertical
      symmetric                <boolean> Determines if wing is symmetric
      origin                   [m]       x, y, and z position
      high_lift                <boolean> Determines if wing is in a high lift configuration
      flaps.                             Flap values are only used if high lift is True
        span_start             [-]       Span start position (.1 is 10% span)
        span_end               [-]       Span end position (.1 is 10% span)
        chord                  [-]       Portion of wing chord used (.1 is 10% chord)

    Outputs:
    wing.
      chords.root              [m]
      chords.tip               [m]
      chords.mean_aerodynamics [m]
      chords.mean_geometric    [m]
      areas.wetted             [m^2]
      areas.affected           [m^2]
      spans.projected          [m]
      aerodynamic_center       [m]      x, y, and z location
      flaps.chord_dimensional  [m]
      flaps.area               [m^2]
        

    Properties Used:
    N/A
    """      
    
    # unpack
    sref        = wing.areas.reference
    taper       = wing.taper
    sweep       = wing.sweeps.quarter_chord
    ar          = wing.aspect_ratio
    t_c_w       = wing.thickness_to_chord
    dihedral    = wing.dihedral 
    vertical    = wing.vertical
    symmetric   = wing.symmetric
    origin      = wing.origin
    
    # calculate
    span       = (ar*sref)**.5
    span_total = span/np.cos(dihedral)
    chord_root = 2*sref/span/(1+taper)
    chord_tip  = taper * chord_root
    mgc        = (chord_root+chord_tip)/2
    
    swet = 2.*span/2.*(chord_root+chord_tip) *  (1.0 + 0.2*t_c_w)

    mac = 2./3.*( chord_root+chord_tip - chord_root*chord_tip/(chord_root+chord_tip) )
    
    # calculate leading edge sweep
    le_sweep = np.arctan( np.tan(sweep) - (4./ar)*(0.-0.25)*(1.-taper)/(1.+taper) )
    
    # estimating aerodynamic center coordinates
    y_coord = span / 6. * (( 1. + 2. * taper ) / (1. + taper))
    x_coord = mac * 0.25 + y_coord * np.tan(le_sweep)
    z_coord = y_coord * np.tan(dihedral)
        
    if vertical:
        temp    = y_coord * 1.
        y_coord = z_coord * 1.
        z_coord = temp

    if symmetric:
        y_coord = 0    
    
    # Total length calculation
    total_length = np.tan(le_sweep)*span/2. + chord_tip
        
    # Computing flap geometry
    affected_area = 0.
    if wing.high_lift:
        flap = wing.control_surfaces.flap
        #compute wing chords at flap start and end
        delta_chord = chord_tip - chord_root
        
        wing_chord_flap_start = chord_root + delta_chord * flap.span_fraction_start 
        wing_chord_flap_end   = chord_root + delta_chord * flap.span_fraction_end
        wing_mac_flap = 2./3.*( wing_chord_flap_start+wing_chord_flap_end - \
                                wing_chord_flap_start*wing_chord_flap_end/  \
                                (wing_chord_flap_start+wing_chord_flap_end) )
        
        flap.chord_dimensional = wing_mac_flap * flap.chord_fraction
        flap_chord_start        = wing_chord_flap_start * flap.chord_fraction
        flap_chord_end          = wing_chord_flap_end * flap.chord_fraction
        flap.area               = (flap_chord_start + flap_chord_end) * (flap.span_fraction_end- flap.span_fraction_start)*span / 2.    
        affected_area           = (wing_chord_flap_start + wing_chord_flap_end) * (flap.span_fraction_end- flap.span_fraction_start)*span / 2.          
         
    # update
    wing.chords.root                = chord_root
    wing.chords.tip                 = chord_tip
    wing.chords.mean_aerodynamic    = mac
    wing.chords.mean_geometric      = mgc
    wing.sweeps.leading_edge        = le_sweep
    wing.areas.wetted               = swet
    wing.areas.affected             = affected_area
    wing.spans.projected            = span
    wing.spans.total                = span_total
    wing.aerodynamic_center         = [x_coord , y_coord, z_coord]
    wing.total_length               = total_length
    
    return wing


# RCAIDE/Compoments/Nacelles/Stack_Nacelle.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core              import Data, Container  
from .Nacelle import Nacelle
  
# ---------------------------------------------------------------------------------------------------------------------- 
#  Stack_Nacelle
# ----------------------------------------------------------------------------------------------------------------------  
class Stack_Nacelle(Nacelle):
    """
    A nacelle design built up from a series of stacked cross-sectional segments, 
    allowing for complex, non-axisymmetric shapes.

    Attributes
    ----------
    tag : str
        Unique identifier for the nacelle component, defaults to 'stack_nacelle'
        
    segments : Container
        Collection of cross-sectional segments defining the nacelle shape, 
        initialized empty

    Notes
    -----
    The stack nacelle approach provides flexibility in designing complex nacelle 
    shapes through the use of multiple cross-sections. This is particularly useful for:
    
    * Non-axisymmetric designs
    * Complex inlet and nozzle geometries
    * Integration with adjacent structures
    
    **Major Assumptions**
    
    * Segments are ordered from front to back
    * Adjacent segments form smooth transitions
    * Cross-sections are defined using super-ellipse parameters
    
    **Definitions**

    'Stack'
        Series of cross-sectional segments arranged longitudinally to define 
        the nacelle shape
        
    'Cross-section'
        Individual segment defining the nacelle shape at a specific longitudinal station

    See Also
    --------
    RCAIDE.Library.Components.Nacelles.Nacelle
        Base nacelle class
    RCAIDE.Library.Components.Nacelles.Segment
        Individual segments used to build the stack
    RCAIDE.Library.Components.Nacelles.Body_of_Revolution_Nacelle
        Alternative nacelle design approach
    """
    
    def __defaults__(self):
        """
        Sets default values for the stack nacelle attributes.
        """      
        self.tag      = 'stack_nacelle'  
        self.segments = Container() 
        
    def append_segment(self, segment):
        """
        Adds a new segment to the nacelle's segment container.

        Parameters
        ----------
        segment : Data
            Cross-sectional segment to be added to the nacelle definition
        """
        # Assert database type
        if not isinstance(segment,RCAIDE.Library.Components.Nacelles.Segments.Segment):
            raise Exception('input component must be of type Segment')

        # Store data
        self.segments.append(segment)

        return  



## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
# fidelity_zero_wake_convergence.py
#
# Created:  Feb 2022, R. Erhard
# Modified: 

from RCAIDE.Library.Methods.Aerodynamics.Common.Lift.BET_calculations import compute_airfoil_aerodynamics,compute_inflow_and_tip_loss
import numpy as np
import scipy as sp

## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def fidelity_zero_wake_convergence(wake,rotor,wake_inputs):
    """
    Wake evaluation is performed using a simplified vortex wake method for Fidelity Zero, 
    following Helmholtz vortex theory.
    
    Assumptions:
    None

    Source:
    Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006
    http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf

    Inputs:
       self         - rotor wake
       rotor        - SUAVE rotor
       wake_inputs.
          Ua        - Axial velocity
          Ut        - Tangential velocity
          r         - radius distribution
       
       
    Outputs:
       va  - axially-induced velocity from rotor wake
       vt  - tangentially-induced velocity from rotor wake
    
    Properties Used:
    None
    
    """        
    
    # Unpack some wake inputs
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na    

    if wake_inputs.use_2d_analysis:
        PSI    = np.ones((ctrl_pts,Nr,Na))
    else:
        PSI    = np.ones((ctrl_pts,Nr))

    PSI_final,infodict,ier,msg = sp.optimize.fsolve(iteration,PSI,args=(wake_inputs,rotor),xtol=rotor.sol_tolerance,full_output = 1,band=(1,0))
    
    if ier!=1:
        print("Rotor BEVW did not converge to a solution (Stall)")
    
    # Calculate the velocities given PSI
    va, vt = va_vt(PSI_final, wake_inputs, rotor)

    
    return va, vt


## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def iteration(PSI, wake_inputs, rotor):
    """
    Computes the BEVW iteration.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """    
    
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    use_2d_analysis = wake_inputs.use_2d_analysis        
    beta            = wake_inputs.twist_distribution
    c               = wake_inputs.chord_distribution
    r               = wake_inputs.radius_distribution
    a               = wake_inputs.speed_of_sounds
    nu              = wake_inputs.dynamic_viscosities
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na

    # Unpack rotor data        
    R            = rotor.tip_radius
    B            = rotor.number_of_blades
    tc           = rotor.thickness_to_chord
    airfoils     = rotor.airfoils
    a_loc        = rotor.airfoil_polar_stations
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))
    
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    vt           = Ut - Wt

    # compute blade airfoil forces and properties
    Cl, Cdval, alpha, alpha_disc,Ma,W,Re,Re_disc = compute_airfoil_aerodynamics(beta,c,r,R,B,Wa,Wt,a,nu,airfoils,a_loc,ctrl_pts,Nr,Na,tc,use_2d_analysis)

    # compute inflow velocity and tip loss factor
    lamdaw, F, piece = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)

    # compute Newton residual on circulation
    Gamma       = vt*(4.*np.pi*r/B)*F*(1.+(4.*lamdaw*R/(np.pi*B*r))*(4.*lamdaw*R/(np.pi*B*r)))**0.5
    Rsquiggly   = Gamma - 0.5*W*c*Cl
    
    return Rsquiggly.flatten()

## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def va_vt(PSI, wake_inputs, rotor):
    """
    Computes the inflow velocities

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """    
    
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))
    
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    va           = Wa - Ua
    vt           = Ut - Wt

    return va, vt

## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def compute_dR_dpsi(PSI,wake_inputs,rotor):
    """
    Computes the analytical derivative for the BEVW iteration.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    beta            = wake_inputs.twist_distribution
    r               = wake_inputs.radius_distribution
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na    

    # Unpack rotor data        
    R        = rotor.tip_radius
    B        = rotor.number_of_blades      
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))    
    
    
    # An analytical derivative for dR_dpsi used in the Newton iteration for the BEVW
    # This was solved symbolically in Matlab and exported
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    
    lamdaw, F, piece = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)
    
    pi          = np.pi
    pi2         = np.pi**2
    BB          = B*B
    BBB         = BB*B
    f_wt_2      = 4*Wt*Wt
    f_wa_2      = 4*Wa*Wa
    arccos_piece = np.arccos(piece)
    Ucospsi     = U*cos_psi
    Usinpsi     = U*sin_psi
    Utcospsi    = Ut*cos_psi
    Uasinpsi    = Ua*sin_psi
    UapUsinpsi  = (Ua + Usinpsi)
    utpUcospsi  = (Ut + Ucospsi)
    utpUcospsi2 = utpUcospsi*utpUcospsi
    UapUsinpsi2 = UapUsinpsi*UapUsinpsi
    dR_dpsi     = ((4.*U*r*arccos_piece*sin_psi*((16.*UapUsinpsi2)/(BB*pi2*f_wt_2) + 1.)**(0.5))/B -
                   (pi*U*(Ua*cos_psi - Ut*sin_psi)*(beta - np.arctan((Wa+Wa)/(Wt+Wt))))/(2.*(f_wt_2 + f_wa_2)**(0.5))
                   + (pi*U*(f_wt_2 +f_wa_2)**(0.5)*(U + Utcospsi  +  Uasinpsi))/(2.*(f_wa_2/(f_wt_2) + 1.)*utpUcospsi2)
                   - (4.*U*piece*((16.*UapUsinpsi2)/(BB*pi2*f_wt_2) + 1.)**(0.5)*(R - r)*(Ut/2. -
                    (Ucospsi)/2.)*(U + Utcospsi + Uasinpsi ))/(f_wa_2*(1. - np.exp(-(B*(Wt+Wt)*(R -
                    r))/(r*(Wa+Wa))))**(0.5)) + (128.*U*r*arccos_piece*(Wa+Wa)*(Ut/2. - (Ucospsi)/2.)*(U +
                    Utcospsi  + Uasinpsi ))/(BBB*pi2*utpUcospsi*utpUcospsi2*((16.*f_wa_2)/(BB*pi2*f_wt_2) + 1.)**(0.5)))

    dR_dpsi[np.isnan(dR_dpsi)] = 0.1
    
    # This needs to be made into a jacobian
    dR_dpsi = dR_dpsi.flatten()
    L       = np.size(PSI)
    jac     = np.eye(L)*dR_dpsi
    
    return jac


# RCAIDE/Library/Methods/Propulsors/Converters/DC_Motor/design_motor.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# python imports  
from scipy.optimize import minimize 

# ----------------------------------------------------------------------------------------------------------------------
#  design motor 
# ----------------------------------------------------------------------------------------------------------------------     
def design_motor(motor):
    ''' Sizes a DC motor to obtain the best combination of speed constant and resistance values
    by sizing the motor for a design RPM value. Note that this design RPM value can be compute
    from design tip mach. The following properties are computed.  
      motor.speed_constant  (float): speed-constant [untiless] 
      motor.resistance      (float): resistance     [ohms]        
    
    Assumptions:
        None 
    
    Source:
        None
    
    Args:
        motor.no_load_current  (float): no-load current  [A]
        motor.nominal_voltage  (float): nominal voltage  [V]
        motor.angular_velocity (float): angular velocity [radians/s]    
        motor.efficiency       (float): efficiency       [unitless]
        motor.design_torque    (float): design torque    [Nm]
       
    Returns:
       None 
    
    '''     
    # design properties of the motor 
    io    = motor.no_load_current
    v     = motor.nominal_voltage
    omeg  = motor.angular_velocity     
    etam  = motor.efficiency 
    Q     = motor.design_torque 
    
    # define optimizer bounds 
    KV_lower_bound  = 0.01
    Res_lower_bound = 0.001
    KV_upper_bound  = 100
    Res_upper_bound = 10
    
    args       = (v , omeg,  etam , Q , io ) 
    hard_cons  = [{'type':'eq', 'fun': hard_constraint_1,'args': args},{'type':'eq', 'fun': hard_constraint_2,'args': args}] 
    slack_cons = [{'type':'eq', 'fun': slack_constraint_1,'args': args},{'type':'eq', 'fun': slack_constraint_2,'args': args}]  
    bnds       = ((KV_lower_bound, KV_upper_bound), (Res_lower_bound , Res_upper_bound)) 
    
    # try hard constraints to find optimum motor parameters
    sol = minimize(objective, [0.5, 0.1], args=(v , omeg,  etam , Q , io) , method='SLSQP', bounds=bnds, tol=1e-6, constraints=hard_cons) 
    
    if sol.success == False:
        # use slack constraints if optimizer fails and motor parameters cannot be found 
        print('\n Optimum motor design failed. Using slack constraints')
        sol = minimize(objective, [0.5, 0.1], args=(v , omeg,  etam , Q , io) , method='SLSQP', bounds=bnds, tol=1e-6, constraints=slack_cons) 
        if sol.success == False:
            assert('\n Slack contraints failed')  
    
    motor.speed_constant   = sol.x[0]
    motor.resistance       = sol.x[1]    
    
    return   
  
# objective function
def objective(x, v , omeg,  etam , Q , io ): 
    return (v - omeg/x[0])/x[1]   

# hard efficiency constraint
def hard_constraint_1(x, v , omeg,  etam , Q , io ): 
    return etam - (1- (io*x[1])/(v - omeg/x[0]))*(omeg/(v*x[0]))   

# hard torque equality constraint
def hard_constraint_2(x, v , omeg,  etam , Q , io ): 
    return ((v - omeg/x[0])/x[1] - io)/x[0] - Q  

# slack efficiency constraint 
def slack_constraint_1(x, v , omeg,  etam , Q , io ): 
    return abs(etam - (1- (io*x[1])/(v - omeg/x[0]))*(omeg/(v*x[0]))) - 0.2

# slack torque equality constraint 
def slack_constraint_2(x, v , omeg,  etam , Q , io ): 
    return  abs(((v - omeg/x[0])/x[1] - io)/x[0] - Q) - 200 


# RCAIDE/Library/Compoments/Energy/Fuel_Tanks/Fuel_Tank.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Library.Components          import Component
from RCAIDE.Library.Methods.Energy.Sources.Fuel_Tanks.append_fuel_tank_conditions import append_fuel_tank_conditions 

# ----------------------------------------------------------------------------------------------------------------------
#  Fuel Tank
# ---------------------------------------------------------------------------------------------------------------------     
class Fuel_Tank(Component):
    """
    Base class for aircraft fuel tank implementations
    
    Attributes
    ----------
    tag : str
        Identifier for the fuel tank (default: 'fuel_tank')
        
    fuel_selector_ratio : float
        Ratio of fuel flow allocation (default: 1.0)
        
    mass_properties.empty_mass : float
        Mass of empty tank structure [kg] (default: 0.0)
        
    secondary_fuel_flow : float
        Secondary fuel flow rate [kg/s] (default: 0.0)
        
    fuel : Component, optional
        Fuel type stored in tank (default: None)

    Notes
    -----
    The fuel tank base class provides common attributes and methods for
    different types of aircraft fuel tanks. It handles basic fuel storage
    and flow management functionality.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Central_Fuel_Tank
        Center section fuel tank
    RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Wing_Fuel_Tank
        Wing-mounted fuel tank
    """
    
    def __defaults__(self):
        """
        Sets default values for fuel tank attributes
        """          
        self.tag                         = 'fuel_tank'
        self.fuel_selector_ratio         = 1.0 
        self.mass_properties.empty_mass  = 0.0   
        self.secondary_fuel_flow         = 0.0
        self.fuel                        = None
         

    def append_operating_conditions(self,segment,fuel_line):  
        """
        Append fuel tank operating conditions for a flight segment
        
        Parameters
        ----------
        segment : Segment
            Flight segment containing state conditions
        fuel_line : Component
            Connected fuel line component
        """
        append_fuel_tank_conditions(self,segment, fuel_line)  
        return                                          


# RCAIDE/Library/Plots/Performance/Common/set_axes.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def set_axes(axes):
    """This sets the axis parameters for all plots

    Assumptions:
    None

    Source:
    None

    Inputs
    axes

    Outputs:
    axes

    Properties Used:
    N/A
    """

    axes.minorticks_on()
    axes.grid(which='major', linestyle='-', linewidth=0.5, color='grey')
    axes.grid(which='minor', linestyle=':', linewidth=0.5, color='grey')
    axes.grid(True)
    axes.get_yaxis().get_major_formatter().set_scientific(False)
    axes.get_yaxis().get_major_formatter().set_useOffset(False)

    return



# RCAIDE/Library/Compoments/Network.py
# 
# Created:  Jul 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Container as ContainerBase
from RCAIDE.Framework.Core import Data
from .Mass_Properties import Mass_Properties

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Network
# ----------------------------------------------------------------------------------------------------------------------        
class Network(Data):
    """
    Base class for component networks that manage connections and interactions between 
    system components.

    Attributes
    ----------
    tag : str
        Unique identifier for the network, defaults to 'Network'
        
    mass_properties : Mass_Properties
        Mass and inertia properties, initialized empty
        
    origin : ndarray
        3D coordinates [x, y, z] defining network's reference point, 
        defaults to [0.0, 0.0, 0.0]
        
    inputs : Data
        Collection of network input parameters, initialized empty
        
    outputs : Data
        Collection of network output parameters, initialized empty

    Notes
    -----
    The Network class serves as a framework for managing interconnected components. 
    It provides:
    
    * Input/output parameter management
    * Mass tracking for network components
    * Spatial positioning capabilities
    * Container functionality for sub-networks

    See Also
    --------
    RCAIDE.Library.Components.Mass_Properties
        Class containing mass and inertia data
    RCAIDE.Framework.Core.Data
        Parent class providing data structure functionality
    """
    def __defaults__(self):
        """
        Sets default values for the network attributes.
        """         
        self.tag             = 'Network' 
        self.mass_properties = Mass_Properties()
        self.origin          = np.array([[0.0,0.0,0.0]]) 
        self.inputs          = Data()
        self.outputs         = Data()
        
# ----------------------------------------------------------------------------------------------------------------------
#  Network Container
# ----------------------------------------------------------------------------------------------------------------------     
class Container(ContainerBase):
    """
    Container class for managing collections of networks.

    Notes
    -----
    The Container class provides organization and mass calculation functionality 
    for groups of networks. Key features include:
    
    * Recursive mass summation
    * Moment calculation about reference points
    * Network hierarchy management

    See Also
    --------
    RCAIDE.Framework.Core.Container
        Parent class providing base container functionality
    """
    pass 
 
    def sum_mass(self):
        """
        Recursively calculates total mass of all contained networks.

        Returns
        -------
        float
            Total mass in kg

        Notes
        -----
        Traverses the network tree and sums mass_properties.mass values from:
        
        * Individual networks
        * Nested network containers (recursive)
        """   
        total = 0.0
        for key,Comp in self.items():
            if isinstance(Comp,Network.Container):
                total += Comp.sum_mass() # recursive!
            elif isinstance(Comp,Network):
                total += Comp.mass_properties.mass
                
        return total
    
    def total_moment(self):
        """
        Recursively calculates total moment of all contained networks.

        Returns
        -------
        ndarray
            Total moment vector about the origin

        Notes
        -----
        Calculates moments considering:
        * Network masses
        * Network origins
        * Network centers of gravity
        * Nested containers (recursive)
        """   
        total = np.array([[0.0,0.0,0.0]])
        for key,Comp in self.items():
            if isinstance(Comp,Network.Container):
                total += Comp.total_moment() # recursive!
            elif isinstance(Comp,Network):
                total += Comp.mass_properties.mass*(np.sum(np.array(Comp.origin),axis=0)/len(Comp.origin)+Comp.mass_properties.center_of_gravity)

        return total
    
# ------------------------------------------------------------
#  Handle Linking
# ------------------------------------------------------------

Network.Container = Container



# RCAIDE/Library/Missions/Common/Residuals/flight_dynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

import RCAIDE
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Residual Total Forces
# ---------------------------------------------------------------------------------------------------------------------- 
def flight_dynamics(segment):
    '''Packs the residuals related to the flight dynamics.
       This includes the force and moment equations about x,y and z
    
    Assumptions:
        N/A
    
    Inputs:
        segment 
        
    Outputs:
        None   
    '''
    transition_seg_flag =  type(segment) == RCAIDE.Framework.Mission.Segments.Transition.Constant_Acceleration_Constant_Angle_Linear_Climb
    ground_seg_flag =  (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Landing) or\
        (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Takeoff) or \
        (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Ground)

    if transition_seg_flag or ground_seg_flag: 
        v       = segment.state.conditions.frames.inertial.velocity_vector
        D       = segment.state.numerics.time.differentiate 
        segment.state.conditions.frames.inertial.acceleration_vector = np.dot(D,v)

    FT = segment.state.conditions.frames.inertial.total_force_vector
    a  = segment.state.conditions.frames.inertial.acceleration_vector    

    if transition_seg_flag: 
        omega = segment.state.conditions.frames.inertial.angular_velocity_vector
        D  = segment.state.numerics.time.differentiate   
        segment.state.conditions.frames.inertial.angular_acceleration_vector =  np.dot(D,omega)         

    MT      = segment.state.conditions.frames.inertial.total_moment_vector    
    ang_acc = segment.state.conditions.frames.inertial.angular_acceleration_vector  
    m       = segment.state.conditions.weights.total_mass
    I       = segment.analyses.aerodynamics.vehicle.mass_properties.moments_of_inertia.tensor  

    if ground_seg_flag:
        vf = segment.velocity_end
        if vf == 0.0: vf = 0.01 
        segment.state.residuals.force_x[:,0] = FT[1:,0]/m[1:,0] - a[1:,0] 
        segment.state.residuals.final_velocity_error = (v[-1,0] - vf)
    else: 
        if segment.flight_dynamics.force_x: 
            segment.state.residuals.force_x[:,0] = FT[:,0]/m[:,0] - a[:,0]  
        if segment.flight_dynamics.force_y: 
            segment.state.residuals.force_y[:,0] = FT[:,1]/m[:,0] - a[:,1]    
        if segment.flight_dynamics.force_z: 
            segment.state.residuals.force_z[:,0] = FT[:,2]/m[:,0] - a[:,2]  
        if  segment.flight_dynamics.moment_x:
            segment.state.residuals.moment_x[:,0] = MT[:,0]/I[0,0] - ang_acc[:,0]   
        if  segment.flight_dynamics.moment_y:
            segment.state.residuals.moment_y[:,0] = MT[:,1]/I[1,1] - ang_acc[:,1]   
        if  segment.flight_dynamics.moment_z:
            segment.state.residuals.moment_z[:,0] = MT[:,2]/I[2,2] - ang_acc[:,2] 
     
    return
